"""Shared test fixtures for denon_rs232."""

import asyncio
from collections.abc import Callable
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

import denon_rs232
import denon_rs232.receiver as denon_receiver
from denon_rs232 import DenonReceiver
from denon_rs232.models import ReceiverModel

# Speed up tests by reducing delays
denon_rs232.COMMAND_TIMEOUT = 0.1
denon_rs232.MULTI_RESPONSE_DELAY = 0.01
denon_rs232.PROBE_TIMEOUT = 0.01
denon_rs232.POWER_ON_DELAY = 0.01
denon_receiver.COMMAND_TIMEOUT = 0.1
denon_receiver.MULTI_RESPONSE_DELAY = 0.01
denon_receiver.PROBE_TIMEOUT = 0.01
denon_receiver.POWER_ON_DELAY = 0.01

# Default responses for all query prefixes during startup.
DEFAULT_QUERY_RESPONSES: dict[str, list[str]] = {
    "PW": ["PWON"],
    "ZM": ["ZMON"],
    "MV": ["MVMAX 98", "MVMIN 99", "MV80"],
    "MU": ["MUOFF"],
    "SI": ["SICD"],
    "MS": ["MSSTEREO"],
    "SD": ["SDAUTO"],
    "SV": ["SVDVD"],
    "SR": ["SRCD"],
    "TF": ["TF105000"],
    "TP": ["TPA1"],
    "CV": ["CVFL 50", "CVFR 50", "CVC 50", "CVSW 50", "CVSL 50", "CVSR 50"],
    "PS": ["PSTONE DEFEAT OFF", "PSSB:OFF", "PSCINEMA EQ.OFF", "PSMODE : CINEMA"],
    "TM": ["TMFM", "TMAUTO"],
    "Z2": ["Z2OFF"],
    "Z3": ["Z3OFF"],
}


class MockSerialConnection:
    """Mock the serial reader/writer pair with auto-response support."""

    def __init__(self):
        self.reader = asyncio.StreamReader()
        self.writer = MagicMock()
        self.writer.write = MagicMock()
        self.writer.drain = AsyncMock()
        self.writer.close = MagicMock()
        self.writer.wait_closed = AsyncMock()
        self.written_data: list[bytes] = []
        self._query_responses: dict[str, list[str]] = {}
        self._command_handler: Callable[[str], None] | None = None
        self.writer.write.side_effect = self._on_write

    def _on_write(self, data: bytes) -> None:
        """Track written data and auto-respond to queries."""
        self.written_data.append(data)
        cmd = data.decode("ascii").rstrip("\r")
        if cmd.endswith("?"):
            prefix = cmd[:-1]
            for resp in self._query_responses.get(prefix, []):
                self.inject_response(resp)
        elif self._command_handler is not None:
            self._command_handler(cmd)

    def inject_response(self, message: str) -> None:
        """Simulate receiver sending a message."""
        self.reader.feed_data(f"{message}\r".encode("ascii"))


@pytest.fixture
async def mock_serial():
    return MockSerialConnection()


@pytest.fixture
async def receiver(mock_serial):
    """Create a connected DenonReceiver with mocked serial."""
    recv = DenonReceiver("/dev/ttyUSB0")
    mock_serial._query_responses = dict(DEFAULT_QUERY_RESPONSES)

    async def fake_open(*args, **kwargs):
        return mock_serial.reader, mock_serial.writer

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        await recv.connect()
        await recv.query_state()

    # Clear auto-responses so tests can inject specific responses manually
    mock_serial._query_responses.clear()

    yield recv

    if recv.connected:
        await recv.disconnect()


async def connect_with_defaults(
    mock: MockSerialConnection, model: ReceiverModel | None = None
) -> DenonReceiver:
    """Helper: connect a receiver with default auto-responses."""
    mock._query_responses = dict(DEFAULT_QUERY_RESPONSES)
    recv = DenonReceiver("/dev/ttyUSB0", model=model)

    async def fake_open(*args, **kwargs):
        return mock.reader, mock.writer

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        await recv.connect()
        await recv.query_state()

    return recv



"""Tests for denon_rs232 source probing."""

import pytest

from denon_rs232 import DenonReceiver, InputSource


async def test_probe_sources(receiver, mock_serial):
    """probe_sources should return only sources the receiver accepts."""
    valid_values = {"CD", "DVD", "TUNER", "TV"}

    def handle_command(cmd):
        if cmd.startswith("SI"):
            source_val = cmd[2:]
            if source_val in valid_values:
                mock_serial.inject_response(f"SI{source_val}")

    mock_serial._command_handler = handle_command

    result = await receiver.probe_sources()

    expected = {InputSource(v) for v in valid_values}
    assert result == expected


async def test_probe_sources_restores_original(receiver, mock_serial):
    """probe_sources should restore the original input source."""
    assert receiver.state.main_zone.input_source == InputSource.CD

    valid_values = {"CD", "DVD"}

    def handle_command(cmd):
        if cmd.startswith("SI"):
            source_val = cmd[2:]
            if source_val in valid_values:
                mock_serial.inject_response(f"SI{source_val}")

    mock_serial._command_handler = handle_command

    await receiver.probe_sources()

    # Last SI command written should be the restore to CD
    si_commands = [
        d
        for d in mock_serial.written_data
        if d.startswith(b"SI") and not d.endswith(b"?\r")
    ]
    assert si_commands[-1] == b"SICD\r"
    # State should be back to CD
    assert receiver.state.main_zone.input_source == InputSource.CD


async def test_probe_includes_current_source(receiver, mock_serial):
    """probe_sources should include the current source even without re-testing it."""
    # After connect, input source is CD. Even if no handler responds,
    # CD should be in the result because it's the current source.
    result = await receiver.probe_sources()

    assert InputSource.CD in result


async def test_probe_disconnected_raises():
    """probe_sources should raise if not connected."""
    recv = DenonReceiver("/dev/ttyUSB0")
    with pytest.raises(ConnectionError, match="Not connected"):
        await recv.probe_sources()



"""Async library to control Denon receivers over RS232 using serialx."""

from .const import (
    BAUD_RATE,
    COMMAND_TIMEOUT,
    CR,
    MAX_VOLUME_DB,
    MIN_VOLUME_DB,
    MULTI_RESPONSE_DELAY,
    POWER_ON_DELAY,
    PROBE_TIMEOUT,
    VOLUME_DB_RANGE,
    ZONE3_PREFIX,
    DigitalInputMode,
    InputSource,
    ModeSetting,
    RoomEQ,
    SurroundBack,
    TunerBand,
    TunerMode,
    _MULTI_RESPONSE_PREFIXES,
    _SINGLE_RESPONSE_PREFIXES,
)
from .players import DenonPlayer, MainPlayer, ZonePlayer
from .protocol import (
    channel_volume_to_param as _channel_volume_to_param,
    parse_channel_volume_param as _parse_channel_volume_param,
    parse_volume_param as _parse_volume_param,
    volume_to_param as _volume_to_param,
)
from .receiver import DenonReceiver, StateCallback
from .state import MainZoneState, ReceiverState, ZoneState

__all__ = [
    "BAUD_RATE",
    "COMMAND_TIMEOUT",
    "CR",
    "DenonPlayer",
    "DenonReceiver",
    "DigitalInputMode",
    "InputSource",
    "MainZoneState",
    "ReceiverState",
    "MAX_VOLUME_DB",
    "MIN_VOLUME_DB",
    "MULTI_RESPONSE_DELAY",
    "MainPlayer",
    "POWER_ON_DELAY",
    "ModeSetting",
    "PROBE_TIMEOUT",
    "RoomEQ",
    "StateCallback",
    "SurroundBack",
    "TunerBand",
    "TunerMode",
    "VOLUME_DB_RANGE",
    "ZONE3_PREFIX",
    "ZonePlayer",
    "ZoneState",
    "_MULTI_RESPONSE_PREFIXES",
    "_SINGLE_RESPONSE_PREFIXES",
    "_channel_volume_to_param",
    "_parse_channel_volume_param",
    "_parse_volume_param",
    "_volume_to_param",
]



"""CLI to test a Denon receiver over RS232.

Usage:
    python -m denon_rs232 /dev/ttyUSB0
    python -m denon_rs232 /dev/ttyUSB0 --probe
    python -m denon_rs232 /dev/ttyUSB0 --zone3-prefix Z1
"""

from __future__ import annotations

import argparse
import asyncio
import sys

from . import DenonReceiver, ReceiverState, ZONE3_PREFIX


def _format_db(db: float | None) -> str:
    if db is None:
        return "?"
    if db == -80.0:
        return "MIN"
    if db >= 0:
        return f"+{db:.1f} dB"
    return f"{db:.1f} dB"


def _format_enum(val: object | None) -> str:
    if val is None:
        return "?"
    if hasattr(val, "value"):
        return val.value
    return str(val)


def _print_state(state: ReceiverState) -> None:
    print()
    print("=== Receiver Status ===")
    print()

    mz = state.main_zone

    print(
        f"  Power:           {'ON' if state.power else 'STANDBY' if state.power is not None else '?'}"
    )
    print(
        f"  Main zone:       {'ON' if mz.power else 'OFF' if mz.power is not None else '?'}"
    )
    print(f"  Volume:          {_format_db(mz.volume)}")
    print(
        f"  Mute:            {'ON' if mz.mute else 'OFF' if mz.mute is not None else '?'}"
    )
    print(f"  Input source:    {_format_enum(mz.input_source)}")
    print(f"  Surround mode:   {mz.surround_mode or '?'}")
    print(f"  Digital input:   {_format_enum(mz.digital_input)}")

    if mz.video_select is not None:
        print(f"  Video select:    {_format_enum(mz.video_select)}")
    if mz.rec_select is not None:
        print(f"  Rec select:      {_format_enum(mz.rec_select)}")

    # Parameter settings
    ps_lines: list[str] = []
    if mz.tone_defeat is not None:
        ps_lines.append(f"Tone defeat {'ON' if mz.tone_defeat else 'OFF'}")
    if mz.surround_back is not None:
        ps_lines.append(f"Surround back: {mz.surround_back.value}")
    if mz.cinema_eq is not None:
        ps_lines.append(f"Cinema EQ {'ON' if mz.cinema_eq else 'OFF'}")
    if mz.mode_setting is not None:
        ps_lines.append(f"Mode: {mz.mode_setting.value}")
    if mz.room_eq is not None:
        ps_lines.append(f"Room EQ: {mz.room_eq.value}")
    if ps_lines:
        print()
        print("  Parameters:")
        for line in ps_lines:
            print(f"    {line}")

    # Channel volumes
    if mz.channel_volumes:
        print()
        print("  Channel volumes:")
        for ch, db in sorted(mz.channel_volumes.items()):
            print(f"    {ch:>3s}:  {_format_db(db)}")

    # Tuner
    if mz.tuner_frequency or mz.tuner_preset:
        print()
        print("  Tuner:")
        if mz.tuner_band:
            print(f"    Band:       {mz.tuner_band.value}")
        if mz.tuner_frequency:
            print(f"    Frequency:  {mz.tuner_frequency}")
        if mz.tuner_preset:
            print(f"    Preset:     {mz.tuner_preset}")
        if mz.tuner_mode:
            print(f"    Mode:       {mz.tuner_mode.value}")

    # Zones
    for label, zone in [("Zone 2", state.zone_2), ("Zone 3", state.zone_3)]:
        if zone.power is not None:
            print()
            print(f"  {label}:")
            print(f"    Power:   {'ON' if zone.power else 'OFF'}")
            if zone.input_source is not None:
                print(f"    Source:  {zone.input_source.value}")
            if zone.volume is not None:
                print(f"    Volume:  {_format_db(zone.volume)}")

    print()


async def _run(port: str, probe: bool, zone3_prefix: str) -> None:
    receiver = DenonReceiver(port, zone3_prefix=zone3_prefix)

    print(f"Connecting to {port}...")
    try:
        await receiver.connect()
        print("Querying receiver state...")
        await receiver.query_state()
    except ConnectionError as err:
        print(f"Error: {err}", file=sys.stderr)
        sys.exit(1)

    try:
        _print_state(receiver.state)

        if probe:
            print("Probing input sources (this will briefly switch inputs)...")
            print()
            sources = await receiver.probe_sources()
            print(f"Available sources ({len(sources)}):")
            for source in sorted(sources, key=lambda s: s.value):
                print(f"  - {source.value}")
            print()
    finally:
        await receiver.disconnect()


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Test a Denon receiver over RS232",
    )
    parser.add_argument("port", help="Serial port (e.g. /dev/ttyUSB0)")
    parser.add_argument(
        "--probe",
        action="store_true",
        help="Probe available input sources",
    )
    parser.add_argument(
        "--zone3-prefix",
        default=ZONE3_PREFIX,
        choices=["Z1", "Z3"],
        help="Zone 3 command prefix (Z1 for AVR-3803/3805, Z3 for modern; default: Z3)",
    )
    args = parser.parse_args()
    asyncio.run(_run(args.port, args.probe, args.zone3_prefix))


if __name__ == "__main__":
    main()



"""Tests for denon_rs232 query, control, and event handling."""

import asyncio
from unittest.mock import AsyncMock, patch

import pytest

from conftest import (
    DEFAULT_QUERY_RESPONSES,
    MockSerialConnection,
    connect_with_defaults,
)

from denon_rs232 import (
    _MULTI_RESPONSE_PREFIXES,
    _SINGLE_RESPONSE_PREFIXES,
    DenonReceiver,
    MainZoneState,
    ReceiverState,
    DigitalInputMode,
    InputSource,
    ModeSetting,
    RoomEQ,
    SurroundBack,
    TunerBand,
    TunerMode,
    _channel_volume_to_param,
    _parse_channel_volume_param,
    _parse_volume_param,
    _volume_to_param,
)
from denon_rs232.models import (
    AVR_3805, 
    AVR_X2700H,
)


# -- Master volume conversion tests --


def test_parse_volume_zero_db():
    assert _parse_volume_param("80") == 0.0


def test_parse_volume_max_18db():
    assert _parse_volume_param("98") == 18.0


def test_parse_volume_min():
    assert _parse_volume_param("99") == -80.0


def test_parse_volume_negative():
    assert _parse_volume_param("60") == -20.0


def test_parse_volume_half_db_positive():
    assert _parse_volume_param("805") == 0.5


def test_parse_volume_half_db_negative():
    assert _parse_volume_param("795") == -0.5


def test_parse_volume_half_db_large():
    assert _parse_volume_param("505") == -29.5


def test_volume_to_param_zero_db():
    assert _volume_to_param(0.0) == "80"


def test_volume_to_param_max():
    assert _volume_to_param(18.0) == "98"


def test_volume_to_param_min():
    assert _volume_to_param(-80.0) == "99"


def test_volume_to_param_negative():
    assert _volume_to_param(-20.0) == "60"


def test_volume_to_param_half_db():
    assert _volume_to_param(0.5) == "805"


def test_volume_to_param_half_db_negative():
    assert _volume_to_param(-0.5) == "795"


def test_volume_to_param_zero_padding():
    assert _volume_to_param(-78) == "02"


def test_volume_to_param_zero_padding_half():
    assert _volume_to_param(-78.5) == "015"


def test_volume_to_param_half_step_on():
    assert _volume_to_param(-30) == "50"
    assert _volume_to_param(-30, half_step=True) == "50"
    assert _volume_to_param(-29.5) == "505"
    assert _volume_to_param(-29.5, half_step=True) == "505"


def test_volume_to_param_half_step_off():
    assert _volume_to_param(-30, half_step=False) == "50"
    assert _volume_to_param(-29.5, half_step=False) == "50"
    assert _volume_to_param(-29, half_step=False) == "51"
    assert _volume_to_param(-28.6, half_step=False) == "51"
    assert _volume_to_param(-28.5, half_step=False) == "51"
    assert _volume_to_param(-28.4, half_step=False) == "51"
    assert _volume_to_param(-0.5, half_step=False) == "79"
    assert _volume_to_param(0.5, half_step=False) == "80"
    assert _volume_to_param(-78.5, half_step=False) == "01"


def test_volume_roundtrip():
    for db in [-80, -20, -10.5, -0.5, 0, 0.5, 10, 18]:
        assert _parse_volume_param(_volume_to_param(db)) == db


# -- Channel volume conversion tests --


def test_parse_channel_volume_zero_db():
    assert _parse_channel_volume_param("50") == 0.0


def test_parse_channel_volume_max():
    assert _parse_channel_volume_param("62") == 12.0


def test_parse_channel_volume_min():
    assert _parse_channel_volume_param("38") == -12.0


def test_parse_channel_volume_off():
    assert _parse_channel_volume_param("00") == -50.0


def test_parse_channel_volume_half_db():
    assert _parse_channel_volume_param("505") == 0.5


def test_channel_volume_to_param_zero():
    assert _channel_volume_to_param(0.0) == "50"


def test_channel_volume_to_param_max():
    assert _channel_volume_to_param(12.0) == "62"


def test_channel_volume_to_param_min():
    assert _channel_volume_to_param(-12.0) == "38"


def test_channel_volume_to_param_half():
    assert _channel_volume_to_param(0.5) == "505"


def test_channel_volume_roundtrip():
    for db in [-12, -6.5, -1, 0, 0.5, 6, 12]:
        assert _parse_channel_volume_param(_channel_volume_to_param(db)) == db


# -- State tests --


def test_initial_state():
    state = ReceiverState()
    assert state.power is None
    assert state.main_zone.power is None
    assert state.main_zone.mute is None
    assert state.main_zone.volume is None
    assert state.main_zone.volume_max is None
    assert state.main_zone.volume_min is None
    assert state.main_zone.input_source is None
    assert state.main_zone.surround_mode is None
    assert state.main_zone.channel_volumes == {}
    assert state.zone_2.power is None
    assert state.zone_3.power is None


def test_state_copy():
    state = ReceiverState(
        power=True,
        main_zone=MainZoneState(
            volume=0.0,
            volume_max=18.0,
            volume_min=-80.0,
        ),
    )
    state.main_zone.channel_volumes["FL"] = 0.0
    state.zone_2.power = True
    copy = state.copy()
    assert copy.power is True
    assert copy.main_zone.volume == 0.0
    assert copy.main_zone.volume_max == 18.0
    assert copy.main_zone.volume_min == -80.0
    assert copy.main_zone.channel_volumes == {"FL": 0.0}
    assert copy.zone_2.power is True
    # Verify deep copy
    copy.power = False
    copy.main_zone.channel_volumes["FL"] = 5.0
    copy.zone_2.power = False
    assert state.power is True
    assert state.main_zone.channel_volumes["FL"] == 0.0
    assert state.zone_2.power is True


# -- Connection tests --


async def test_connect_verifies_power(mock_serial):
    recv = DenonReceiver("/dev/ttyUSB0")

    async def fake_open(*args, **kwargs):
        return mock_serial.reader, mock_serial.writer

    mock_serial._query_responses = dict(DEFAULT_QUERY_RESPONSES)

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        await recv.connect()

    assert recv.state.power is True
    sent_queries = [data.decode("ascii") for data in mock_serial.written_data]
    assert sent_queries == ["PW?\r"]

    await recv.disconnect()


async def test_query_state_populates_full_state(receiver):
    """query_state() should populate state from all startup queries."""
    state = receiver.state
    assert state.power is True
    assert state.main_zone.power is True
    assert state.main_zone.volume == 0.0
    assert state.main_zone.volume_max == 18.0
    assert state.main_zone.volume_min == -80.0
    assert state.main_zone.mute is False
    assert state.main_zone.input_source == InputSource.CD
    assert state.main_zone.surround_mode == "STEREO"
    assert state.main_zone.digital_input == DigitalInputMode.AUTO
    assert state.main_zone.video_select == InputSource.DVD
    assert state.main_zone.rec_select == InputSource.CD
    assert state.main_zone.tuner_frequency == "105000"
    assert state.main_zone.tuner_preset == "A1"
    assert state.main_zone.tuner_band == TunerBand.FM
    assert state.main_zone.tuner_mode == TunerMode.AUTO
    assert state.main_zone.tone_defeat is False
    assert state.main_zone.surround_back == SurroundBack.OFF
    assert state.main_zone.cinema_eq is False
    assert state.main_zone.mode_setting == ModeSetting.CINEMA
    assert state.main_zone.channel_volumes["FL"] == 0.0
    assert state.main_zone.channel_volumes["FR"] == 0.0
    assert state.zone_2.power is False
    assert state.zone_3.power is False


async def test_receiver_exposes_shared_player_objects(receiver):
    """Receiver should expose live player objects for main and zones."""
    assert receiver.power is True
    assert receiver.main.power is True
    assert receiver.main.input_source == InputSource.CD
    assert receiver.main.volume == 0.0
    assert receiver.main.volume_min == -80.0
    assert receiver.main.volume_max == 18.0
    assert receiver.main.mute is False
    assert receiver.zone_2.power is False
    assert receiver.zone_3.power is False


async def test_query_state_queries_all_prefixes(mock_serial):
    """query_state() should send ? queries for all startup prefixes."""
    recv = await connect_with_defaults(mock_serial)

    # Check all expected query commands were sent.
    # Z1 in _MULTI_RESPONSE_PREFIXES is replaced by the configured zone3 prefix (Z3).
    expected = set(_SINGLE_RESPONSE_PREFIXES) | {
        "Z3" if p == "Z1" else p for p in _MULTI_RESPONSE_PREFIXES
    }
    sent_queries = {
        data[:-1].decode("ascii").replace("?", "")
        for data in mock_serial.written_data
        if data.endswith(b"?\r")
    }
    assert expected == sent_queries

    await recv.disconnect()


async def test_query_state_notifies_subscribers_once(mock_serial):
    """query_state() should fire a single notification despite many queries."""
    recv = DenonReceiver("/dev/ttyUSB0")
    mock_serial._query_responses = dict(DEFAULT_QUERY_RESPONSES)

    async def fake_open(*args, **kwargs):
        return mock_serial.reader, mock_serial.writer

    notifications = []

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        await recv.connect()
        recv.subscribe(notifications.append)
        await recv.query_state()

    assert len(notifications) == 1
    assert notifications[0].main_zone.input_source == InputSource.CD

    await recv.disconnect()


async def test_query_state_no_notification_when_unchanged(receiver, mock_serial):
    """A repeat query_state() with identical state should not notify."""
    mock_serial._query_responses = dict(DEFAULT_QUERY_RESPONSES)

    notifications = []
    receiver.subscribe(notifications.append)

    await receiver.query_state()

    assert notifications == []


async def test_query_state_skips_known_unsupported_model_queries(mock_serial):
    """Known model capabilities should trim unsupported startup probes."""
    recv = await connect_with_defaults(mock_serial, model=AVR_X2700H)

    sent_queries = {
        data[:-1].decode("ascii").replace("?", "")
        for data in mock_serial.written_data
        if data.endswith(b"?\r")
    }

    assert "SR" not in sent_queries
    assert "TF" not in sent_queries
    assert "TP" not in sent_queries
    assert "Z3" not in sent_queries

    state = recv.state
    assert state.main_zone.rec_select is None
    assert state.main_zone.tuner_frequency is None
    assert state.main_zone.tuner_preset is None
    assert state.zone_3.power is None

    await recv.disconnect()


async def test_query_state_stops_at_first_unanswered_query_in_standby(mock_serial):
    """A receiver in standby should not be waited out query after query."""
    mock_serial._query_responses = {"PW": ["PWSTANDBY"], "ZM": ["ZMOFF"]}
    recv = DenonReceiver("/dev/ttyUSB0")

    async def fake_open(*args, **kwargs):
        return mock_serial.reader, mock_serial.writer

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        await recv.connect()
        await recv.query_state()

    sent_queries = {
        data[:-1].decode("ascii").replace("?", "")
        for data in mock_serial.written_data
        if data.endswith(b"?\r")
    }
    multi = {"Z3" if p == "Z1" else p for p in _MULTI_RESPONSE_PREFIXES}

    # ZM answers, MV does not, so the queries after it are not attempted. The
    # multi-response queries are not waited on and are always sent.
    assert sent_queries == {"PW", "ZM", "MV"} | multi
    assert recv.power is False
    assert recv.main.power is False

    await recv.disconnect()


async def test_query_state_in_standby_continues_while_answered(mock_serial):
    """A receiver that keeps answering in standby should be queried in full."""
    mock_serial._query_responses = dict(DEFAULT_QUERY_RESPONSES) | {
        "PW": ["PWSTANDBY"]
    }
    recv = DenonReceiver("/dev/ttyUSB0")

    async def fake_open(*args, **kwargs):
        return mock_serial.reader, mock_serial.writer

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        await recv.connect()
        await recv.query_state()

    sent_queries = {
        data[:-1].decode("ascii").replace("?", "")
        for data in mock_serial.written_data
        if data.endswith(b"?\r")
    }
    expected = set(_SINGLE_RESPONSE_PREFIXES) | {
        "Z3" if p == "Z1" else p for p in _MULTI_RESPONSE_PREFIXES
    }

    assert sent_queries == expected
    assert recv.power is False
    assert recv.main.input_source == InputSource.CD
    assert recv.zone_2.power is False

    await recv.disconnect()


async def test_query_state_rechecks_power_before_skipping(mock_serial):
    """A receiver turned on since connect() should still be fully queried."""
    mock_serial._query_responses = {"PW": ["PWSTANDBY"]}
    recv = DenonReceiver("/dev/ttyUSB0")

    async def fake_open(*args, **kwargs):
        return mock_serial.reader, mock_serial.writer

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        await recv.connect()
        assert recv.power is False

        mock_serial._query_responses = dict(DEFAULT_QUERY_RESPONSES)
        await recv.query_state()

    sent_queries = {
        data[:-1].decode("ascii").replace("?", "")
        for data in mock_serial.written_data
        if data.endswith(b"?\r")
    }
    expected = set(_SINGLE_RESPONSE_PREFIXES) | {
        "Z3" if p == "Z1" else p for p in _MULTI_RESPONSE_PREFIXES
    }

    assert sent_queries == expected
    assert recv.power is True
    assert recv.main.input_source == InputSource.CD

    await recv.disconnect()


async def test_power_on_event_queries_withheld_state(mock_serial):
    """A receiver leaving standby should be queried for what it withheld."""
    mock_serial._query_responses = {"PW": ["PWSTANDBY"], "ZM": ["ZMOFF"]}
    recv = DenonReceiver("/dev/ttyUSB0")

    async def fake_open(*args, **kwargs):
        return mock_serial.reader, mock_serial.writer

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        await recv.connect()
        await recv.query_state()

        mock_serial.written_data.clear()
        mock_serial._query_responses = dict(DEFAULT_QUERY_RESPONSES)
        mock_serial.inject_response("PWON")
        await asyncio.sleep(0)

        task = recv._power_on_task
        assert task is not None
        await task

    sent_queries = {
        data[:-1].decode("ascii").replace("?", "")
        for data in mock_serial.written_data
        if data.endswith(b"?\r")
    }
    expected = set(_SINGLE_RESPONSE_PREFIXES) | {
        "Z3" if p == "Z1" else p for p in _MULTI_RESPONSE_PREFIXES
    }

    assert sent_queries == expected
    assert recv.main.input_source == InputSource.CD
    assert recv.zone_2.power is False

    await recv.disconnect()


async def test_power_query_at_startup_does_not_schedule_query(mock_serial):
    """Querying a receiver that is already on should not trigger a re-query."""
    recv = await connect_with_defaults(mock_serial)

    assert recv._power_on_task is None

    await recv.disconnect()


async def test_connect_timeout_raises():
    recv = DenonReceiver("/dev/ttyUSB0")
    mock = MockSerialConnection()
    # No auto-responses: power query will timeout

    async def fake_open(*args, **kwargs):
        return mock.reader, mock.writer

    with patch(
        "denon_rs232.receiver.serialx.open_serial_connection",
        side_effect=fake_open,
    ):
        with pytest.raises(ConnectionError, match="No response"):
            await recv.connect()


async def test_disconnect(receiver):
    await receiver.disconnect()
    assert not receiver.connected


async def test_written_command_format(receiver, mock_serial):
    assert b"PW?\r" in mock_serial.written_data


# -- Power & main zone command tests --


async def test_power_on(receiver, mock_serial):
    await receiver.power_on()
    assert b"PWON\r" in mock_serial.written_data


async def test_power_standby(receiver, mock_serial):
    await receiver.power_standby()
    assert b"PWSTANDBY\r" in mock_serial.written_data


async def test_main_player_power_on(receiver, mock_serial):
    await receiver.main.power_on()
    assert b"ZMON\r" in mock_serial.written_data


async def test_main_player_power_standby(receiver, mock_serial):
    await receiver.main.power_standby()
    assert b"ZMOFF\r" in mock_serial.written_data


# -- Master volume command tests --


async def test_volume_up(receiver, mock_serial):
    await receiver.main.volume_up()
    assert b"MVUP\r" in mock_serial.written_data


async def test_volume_down(receiver, mock_serial):
    await receiver.main.volume_down()
    assert b"MVDOWN\r" in mock_serial.written_data


async def test_set_volume(receiver, mock_serial):
    await receiver.main.set_volume(0.0)
    assert b"MV80\r" in mock_serial.written_data


async def test_set_volume_half_db(receiver, mock_serial):
    await receiver.main.set_volume(-10.5)
    assert b"MV695\r" in mock_serial.written_data


# -- Channel volume command tests --


async def test_channel_volume_up(receiver, mock_serial):
    await receiver.main.channel_volume_up("FL")
    assert b"CVFL UP\r" in mock_serial.written_data


async def test_channel_volume_down(receiver, mock_serial):
    await receiver.main.channel_volume_down("C")
    assert b"CVC DOWN\r" in mock_serial.written_data


async def test_set_channel_volume(receiver, mock_serial):
    await receiver.main.set_channel_volume("SW", 3.0)
    assert b"CVSW 53\r" in mock_serial.written_data


async def test_set_channel_volume_half_db(receiver, mock_serial):
    await receiver.main.set_channel_volume("FR", -2.5)
    assert b"CVFR 475\r" in mock_serial.written_data


# -- Mute command tests --


async def test_mute_on(receiver, mock_serial):
    await receiver.main.mute_on()
    assert b"MUON\r" in mock_serial.written_data


async def test_mute_off(receiver, mock_serial):
    await receiver.main.mute_off()
    assert b"MUOFF\r" in mock_serial.written_data


# -- Input source command tests --


async def test_select_input_source(receiver, mock_serial):
    await receiver.main.select_input_source(InputSource.DVD)
    assert b"SIDVD\r" in mock_serial.written_data


async def test_select_input_source_with_slash(receiver, mock_serial):
    await receiver.main.select_input_source(InputSource.DBS_SAT)
    assert b"SIDBS/SAT\r" in mock_serial.written_data


# -- Surround mode command tests --


async def test_set_surround_mode(receiver, mock_serial):
    await receiver.main.set_surround_mode("STEREO")
    assert b"MSSTEREO\r" in mock_serial.written_data


async def test_set_surround_mode_with_spaces(receiver, mock_serial):
    await receiver.main.set_surround_mode("DOLBY PRO LOGIC")
    assert b"MSDOLBY PRO LOGIC\r" in mock_serial.written_data


# -- Parameter settings command tests --


async def test_tone_defeat_on(receiver, mock_serial):
    await receiver.main.tone_defeat_on()
    assert b"PSTONE DEFEAT ON\r" in mock_serial.written_data


async def test_tone_defeat_off(receiver, mock_serial):
    await receiver.main.tone_defeat_off()
    assert b"PSTONE DEFEAT OFF\r" in mock_serial.written_data


async def test_set_surround_back(receiver, mock_serial):
    await receiver.main.set_surround_back(SurroundBack.MTRX_ON)
    assert b"PSSB:MTRX ON\r" in mock_serial.written_data


async def test_cinema_eq_on(receiver, mock_serial):
    await receiver.main.cinema_eq_on()
    assert b"PSCINEMA EQ.ON\r" in mock_serial.written_data


async def test_cinema_eq_off(receiver, mock_serial):
    await receiver.main.cinema_eq_off()
    assert b"PSCINEMA EQ.OFF\r" in mock_serial.written_data


async def test_set_mode_setting(receiver, mock_serial):
    await receiver.main.set_mode_setting(ModeSetting.MUSIC)
    assert b"PSMODE : MUSIC\r" in mock_serial.written_data


async def test_set_room_eq(receiver, mock_serial):
    await receiver.main.set_room_eq(RoomEQ.FLAT)
    assert b"PSROOM EQ:FLAT\r" in mock_serial.written_data


# -- Digital input command tests --


async def test_set_digital_input(receiver, mock_serial):
    await receiver.main.set_digital_input(DigitalInputMode.AUTO)
    assert b"SDAUTO\r" in mock_serial.written_data


async def test_set_digital_input_ext(receiver, mock_serial):
    await receiver.main.set_digital_input(DigitalInputMode.EXT_IN_1)
    assert b"SDEXT.IN-1\r" in mock_serial.written_data


# -- Video select command tests --


async def test_set_video_select(receiver, mock_serial):
    await receiver.main.set_video_select(InputSource.DVD)
    assert b"SVDVD\r" in mock_serial.written_data


async def test_cancel_video_select(receiver, mock_serial):
    await receiver.main.cancel_video_select()
    assert b"SVSOURCE\r" in mock_serial.written_data


# -- Rec select command tests --


async def test_set_rec_select(receiver, mock_serial):
    await receiver.main.set_rec_select(InputSource.CD)
    assert b"SRCD\r" in mock_serial.written_data


async def test_cancel_rec_select(receiver, mock_serial):
    await receiver.main.cancel_rec_select()
    assert b"SRSOURCE\r" in mock_serial.written_data


# -- Tuner command tests --


async def test_tuner_frequency_up(receiver, mock_serial):
    await receiver.main.tuner_frequency_up()
    assert b"TFUP\r" in mock_serial.written_data


async def test_tuner_frequency_down(receiver, mock_serial):
    await receiver.main.tuner_frequency_down()
    assert b"TFDOWN\r" in mock_serial.written_data


async def test_set_tuner_frequency(receiver, mock_serial):
    await receiver.main.set_tuner_frequency("105000")
    assert b"TF105000\r" in mock_serial.written_data


async def test_tuner_preset_up(receiver, mock_serial):
    await receiver.main.tuner_preset_up()
    assert b"TPUP\r" in mock_serial.written_data


async def test_tuner_preset_down(receiver, mock_serial):
    await receiver.main.tuner_preset_down()
    assert b"TPDOWN\r" in mock_serial.written_data


async def test_set_tuner_preset(receiver, mock_serial):
    await receiver.main.set_tuner_preset("A1")
    assert b"TPA1\r" in mock_serial.written_data


async def test_set_tuner_band(receiver, mock_serial):
    await receiver.main.set_tuner_band(TunerBand.FM)
    assert b"TMFM\r" in mock_serial.written_data


async def test_set_tuner_mode(receiver, mock_serial):
    await receiver.main.set_tuner_mode(TunerMode.AUTO)
    assert b"TMAUTO\r" in mock_serial.written_data


# -- Zone 2 command tests --


async def test_zone2_power_on(receiver, mock_serial):
    await receiver.zone_2.power_on()
    assert b"Z2ON\r" in mock_serial.written_data


async def test_zone2_player_power_on(receiver, mock_serial):
    await receiver.zone_2.power_on()
    assert b"Z2ON\r" in mock_serial.written_data


async def test_zone2_power_standby(receiver, mock_serial):
    await receiver.zone_2.power_standby()
    assert b"Z2OFF\r" in mock_serial.written_data


async def test_zone2_select_input_source(receiver, mock_serial):
    await receiver.zone_2.select_input_source(InputSource.CD)
    assert b"Z2CD\r" in mock_serial.written_data


async def test_zone2_volume_up(receiver, mock_serial):
    await receiver.zone_2.volume_up()
    assert b"Z2UP\r" in mock_serial.written_data


async def test_zone2_set_volume(receiver, mock_serial):
    await receiver.zone_2.set_volume(0.0)
    assert b"Z280\r" in mock_serial.written_data


async def test_zone2_set_volume_no_half_db_steps(mock_serial):
    """Receiver without support for setting the zone volume in half dB steps
       shall only issue four character short zone commands."""
    recv = await connect_with_defaults(mock_serial, model=AVR_3805)

    await recv.zone_2.set_volume(-10.5)
    assert b"Z269\r" in mock_serial.written_data

    await recv.disconnect()


# -- Zone 3 command tests (default Z3 prefix) --


async def test_zone3_power_on(receiver, mock_serial):
    await receiver.zone_3.power_on()
    assert b"Z3ON\r" in mock_serial.written_data


async def test_zone3_player_power_on(receiver, mock_serial):
    await receiver.zone_3.power_on()
    assert b"Z3ON\r" in mock_serial.written_data


async def test_zone3_power_standby(receiver, mock_serial):
    await receiver.zone_3.power_standby()
    assert b"Z3OFF\r" in mock_serial.written_data


async def test_zone3_select_input_source(receiver, mock_serial):
    await receiver.zone_3.select_input_source(InputSource.TUNER)
    assert b"Z3TUNER\r" in mock_serial.written_data


async def test_zone3_set_volume(receiver, mock_serial):
    await receiver.zone_3.set_volume(-10.0)
    assert b"Z370\r" in mock_serial.written_data


async def test_zone3_set_volume_no_half_db_steps(mock_serial):
    """Receiver without support for setting the zone volume in half dB steps
       shall only issue four character long zone commands."""
    recv = await connect_with_defaults(mock_serial, model=AVR_3805)

    await recv.zone_3.set_volume(-10.5)
    assert b"Z169\r" in mock_serial.written_data

    await recv.disconnect()


# -- Query tests --


async def test_query_power(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("PWSTANDBY")

    task = asyncio.create_task(respond())
    result = await receiver.query_power()
    await task
    assert result is False


async def test_query_volume(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("MVMAX 98")
        mock_serial.inject_response("MVMIN 99")
        mock_serial.inject_response("MV75")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_volume()
    await task
    assert result == -5.0
    assert receiver.state.main_zone.volume_max == 18.0
    assert receiver.state.main_zone.volume_min == -80.0


async def test_query_mute(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("MUOFF")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_mute()
    await task
    assert result is False


async def test_query_main_power(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("ZMON")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_power()
    await task
    assert result is True


async def test_query_input_source(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("SIDVD")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_input_source()
    await task
    assert result == InputSource.DVD


async def test_query_surround_mode(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("MSDOLBY PL2X C")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_surround_mode()
    await task
    assert result == "DOLBY PL2X C"


async def test_query_digital_input(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("SDAUTO")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_digital_input()
    await task
    assert result == DigitalInputMode.AUTO


async def test_query_digital_input_no_returns_none(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("SDNO")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_digital_input()
    await task
    assert result is None


async def test_query_video_select(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("SVDVD")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_video_select()
    await task
    assert result == InputSource.DVD


async def test_query_video_select_off_returns_none(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("SVOFF")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_video_select()
    await task
    assert result is None


async def test_query_rec_select(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("SRCD")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_rec_select()
    await task
    assert result == InputSource.CD


async def test_query_tuner_frequency(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("TF105000")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_tuner_frequency()
    await task
    assert result == "105000"


async def test_query_tuner_preset(receiver, mock_serial):
    async def respond():
        await asyncio.sleep(0.05)
        mock_serial.inject_response("TPA1")

    task = asyncio.create_task(respond())
    result = await receiver.main.query_tuner_preset()
    await task
    assert result == "A1"


# -- Event tests: input source --


async def test_input_source_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SIDVD")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.input_source == InputSource.DVD


async def test_input_source_event_with_slash(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SIDBS/SAT")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.input_source == InputSource.DBS_SAT


# -- Event tests: surround mode --


async def test_surround_mode_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("MSDIRECT")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.surround_mode == "DIRECT"


async def test_surround_mode_event_combined(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("MSDOLBY D+PL2X C")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.surround_mode == "DOLBY D+PL2X C"


# -- Event tests: channel volume --


async def test_channel_volume_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("CVFL 52")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.channel_volumes["FL"] == 2.0


async def test_channel_volume_event_subwoofer(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("CVSW 55")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.channel_volumes["SW"] == 5.0


async def test_channel_volume_event_surround_back_left(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("CVSBL 48")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.channel_volumes["SBL"] == -2.0


async def test_channel_volume_event_accepts_unknown_channel(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("CVFH 51")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.channel_volumes["FH"] == 1.0


async def test_channel_volume_up_event_no_state_change(receiver, mock_serial):
    """CV UP/DOWN events should not update state."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("CVFL UP")
    await asyncio.sleep(0.1)

    assert len(states) == 0


async def test_duplicate_channel_volume_event_no_state_change(receiver, mock_serial):
    """Duplicate channel volume values should not notify subscribers."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("CVFL 50")
    await asyncio.sleep(0.1)

    assert len(states) == 0


async def test_multiple_channel_volume_events(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("CVFL 50")
    mock_serial.inject_response("CVFR 52")
    mock_serial.inject_response("CVC 48")
    await asyncio.sleep(0.1)

    assert len(states) == 2
    assert states[-1].main_zone.channel_volumes["FL"] == 0.0
    assert states[-1].main_zone.channel_volumes["FR"] == 2.0
    assert states[-1].main_zone.channel_volumes["C"] == -2.0


# -- Event tests: parameter settings --


async def test_tone_defeat_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PSTONE DEFEAT ON")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.tone_defeat is True


async def test_tone_defeat_off_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PSTONE DEFEAT ON")
    await asyncio.sleep(0.05)
    mock_serial.inject_response("PSTONE DEFEAT OFF")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.tone_defeat is False


async def test_surround_back_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PSSB:MTRX ON")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.surround_back == SurroundBack.MTRX_ON


async def test_surround_back_off_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PSSB:MTRX ON")
    await asyncio.sleep(0.05)
    mock_serial.inject_response("PSSB:OFF")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.surround_back == SurroundBack.OFF


async def test_cinema_eq_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PSCINEMA EQ.ON")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.cinema_eq is True


async def test_mode_setting_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PSMODE : MUSIC")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.mode_setting == ModeSetting.MUSIC


async def test_room_eq_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PSROOM EQ:FLAT")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.room_eq == RoomEQ.FLAT


# -- Event tests: digital input --


async def test_digital_input_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SDANALOG")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.digital_input == DigitalInputMode.ANALOG


async def test_digital_input_no_event(receiver, mock_serial, caplog):
    """SD NO should clear digital_input to None without warning."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SDNO")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.digital_input is None
    assert "Unknown digital input mode: NO" not in caplog.text


# -- Event tests: video select --


async def test_video_select_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SVCD")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.video_select == InputSource.CD


async def test_video_select_source_event(receiver, mock_serial):
    """SV SOURCE should clear video_select to None."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SVSOURCE")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.video_select is None


async def test_video_select_off_event(receiver, mock_serial, caplog):
    """SV OFF should clear video_select to None without warning."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SVOFF")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.video_select is None
    assert "Unknown video source: OFF" not in caplog.text


# -- Event tests: rec select --


async def test_rec_select_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SRCDR/TAPE1")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.rec_select == InputSource.CDR_TAPE1


async def test_rec_select_source_event(receiver, mock_serial):
    """SR SOURCE should clear rec_select to None."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("SRSOURCE")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.rec_select is None


# -- Event tests: tuner --


async def test_tuner_frequency_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("TF106000")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.tuner_frequency == "106000"


async def test_tuner_frequency_up_no_state(receiver, mock_serial):
    """TF UP/DOWN should not update state."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("TFUP")
    await asyncio.sleep(0.1)

    assert len(states) == 0


async def test_tuner_preset_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("TPB2")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.tuner_preset == "B2"


async def test_tuner_band_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("TMAM")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.tuner_band == TunerBand.AM


async def test_tuner_mode_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("TMMANUAL")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.tuner_mode == TunerMode.MANUAL


# -- Event tests: zone 2 --


async def test_zone2_power_on_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z2ON")
    await asyncio.sleep(0.1)

    assert states[-1].zone_2.power is True


async def test_zone2_power_off_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z2ON")
    await asyncio.sleep(0.05)
    mock_serial.inject_response("Z2OFF")
    await asyncio.sleep(0.1)

    assert states[-1].zone_2.power is False


async def test_zone2_source_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z2DVD")
    await asyncio.sleep(0.1)

    assert states[-1].zone_2.input_source == InputSource.DVD


async def test_zone2_volume_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z280")
    await asyncio.sleep(0.1)

    assert states[-1].zone_2.volume == 0.0


async def test_zone2_volume_up_no_state(receiver, mock_serial):
    """Z2 UP/DOWN should not trigger state change."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z2UP")
    await asyncio.sleep(0.1)

    assert len(states) == 0


async def test_zone2_source_cancel_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    # Set a source first
    mock_serial.inject_response("Z2DVD")
    await asyncio.sleep(0.05)
    # Cancel it
    mock_serial.inject_response("Z2SOURCE")
    await asyncio.sleep(0.1)

    assert states[-1].zone_2.input_source is None


# -- Event tests: zone 3 (both Z3 and Z1 prefixes) --


async def test_zone3_power_on_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z3ON")
    await asyncio.sleep(0.1)

    assert states[-1].zone_3.power is True


async def test_zone3_source_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z3TUNER")
    await asyncio.sleep(0.1)

    assert states[-1].zone_3.input_source == InputSource.TUNER


async def test_zone3_sleep_timer_event_ignored(receiver, mock_serial, caplog):
    """Zone sleep timer payloads should not be treated as sources."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z3SLPOFF")
    await asyncio.sleep(0.1)

    assert len(states) == 0
    assert "Unknown zone source: SLPOFF" not in caplog.text


async def test_zone3_volume_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z370")
    await asyncio.sleep(0.1)

    assert states[-1].zone_3.volume == -10.0


async def test_zone3_legacy_z1_prefix_event(receiver, mock_serial):
    """Z1 prefix (legacy AVR-3803/3805) should also update zone3 state."""
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("Z1ON")
    await asyncio.sleep(0.1)

    assert states[-1].zone_3.power is True


# -- Existing event/subscriber tests --


async def test_subscribe_receives_events(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("MV75")
    await asyncio.sleep(0.1)

    assert len(states) == 1
    assert states[0].main_zone.volume == -5.0


async def test_unsubscribe(receiver, mock_serial):
    states: list[ReceiverState] = []
    unsub = receiver.subscribe(lambda s: states.append(s))
    unsub()

    mock_serial.inject_response("MV75")
    await asyncio.sleep(0.1)

    assert len(states) == 0


async def test_power_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PWSTANDBY")
    await asyncio.sleep(0.1)

    assert states[-1].power is False


async def test_duplicate_power_event_no_state_change(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PWON")
    await asyncio.sleep(0.1)

    assert len(states) == 0


async def test_mute_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("MUON")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.mute is True


async def test_main_zone_event(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("ZMOFF")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.power is False


async def test_max_volume_event(receiver, mock_serial):
    """MVMAX messages should update the stored max volume."""
    states: list[ReceiverState] = []
    receiver._state.main_zone.volume_max = None
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("MVMAX 98")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.volume_max == 18.0
    assert states[-1].main_zone.volume == 0.0


async def test_min_volume_event(receiver, mock_serial):
    """MVMIN messages should update the stored min volume."""
    states: list[ReceiverState] = []
    receiver._state.main_zone.volume_min = None
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("MVMIN 99")
    await asyncio.sleep(0.1)

    assert states[-1].main_zone.volume_min == -80.0
    assert states[-1].main_zone.volume == 0.0


async def test_multiple_events(receiver, mock_serial):
    states: list[ReceiverState] = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.inject_response("PWSTANDBY")
    mock_serial.inject_response("MV75")
    mock_serial.inject_response("MUON")
    await asyncio.sleep(0.1)

    assert len(states) == 3
    assert states[-1].power is False
    assert states[-1].main_zone.volume == -5.0
    assert states[-1].main_zone.mute is True


async def test_bad_callback_doesnt_break(receiver, mock_serial):
    """An exception in a subscriber shouldn't prevent other processing."""

    def bad_callback(state):
        raise RuntimeError("oops")

    good_states: list[ReceiverState] = []
    receiver.subscribe(bad_callback)
    receiver.subscribe(lambda s: good_states.append(s))

    mock_serial.inject_response("PWSTANDBY")
    await asyncio.sleep(0.1)

    assert len(good_states) == 1


# -- Teardown tests --


async def test_read_error_closes_writer():
    """A read error should close the writer and mark disconnected."""
    mock = MockSerialConnection()
    recv = await connect_with_defaults(mock)

    assert recv.connected

    mock.reader.set_exception(OSError("device unplugged"))
    await asyncio.sleep(0.1)

    assert not recv.connected
    mock.writer.close.assert_called_once()
    mock.writer.wait_closed.assert_called()


async def test_read_eof_closes_writer():
    """EOF on the reader should close the writer and mark disconnected."""
    mock = MockSerialConnection()
    recv = await connect_with_defaults(mock)

    assert recv.connected

    mock.reader.feed_eof()
    await asyncio.sleep(0.1)

    assert not recv.connected
    mock.writer.close.assert_called_once()
    mock.writer.wait_closed.assert_called()


async def test_write_error_closes_reader():
    """A write error should tear down the connection including the read loop."""
    mock = MockSerialConnection()
    recv = await connect_with_defaults(mock)

    assert recv.connected
    read_task = recv._read_task

    mock.writer.drain = AsyncMock(side_effect=OSError("device unplugged"))

    with pytest.raises(OSError, match="device unplugged"):
        await recv.power_on()

    assert not recv.connected
    mock.writer.close.assert_called_once()
    assert read_task.cancelled() or read_task.done()


async def test_query_write_error_closes_reader():
    """A write error during a query should tear down the connection."""
    mock = MockSerialConnection()
    recv = await connect_with_defaults(mock)

    assert recv.connected

    mock.writer.drain = AsyncMock(side_effect=OSError("device unplugged"))

    with pytest.raises(OSError, match="device unplugged"):
        await recv.query_power()

    assert not recv.connected
    mock.writer.close.assert_called_once()
    assert len(recv._pending_queries) == 0


async def test_read_error_notifies_none(receiver, mock_serial):
    """Subscribers receive None when the connection is lost via read error."""
    states = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.reader.set_exception(OSError("device unplugged"))
    await asyncio.sleep(0.1)

    assert states[-1] is None


async def test_read_eof_notifies_none(receiver, mock_serial):
    """Subscribers receive None when the connection is lost via EOF."""
    states = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.reader.feed_eof()
    await asyncio.sleep(0.1)

    assert states[-1] is None


async def test_write_error_notifies_none(receiver, mock_serial):
    """Subscribers receive None when the connection is lost via write error."""
    states = []
    receiver.subscribe(lambda s: states.append(s))

    mock_serial.writer.drain = AsyncMock(side_effect=OSError("device unplugged"))

    with pytest.raises(OSError):
        await receiver.power_on()

    assert states[-1] is None


async def test_disconnect_notifies_none(receiver, mock_serial):
    """Subscribers receive None on explicit disconnect."""
    states = []
    receiver.subscribe(lambda s: states.append(s))

    await receiver.disconnect()

    assert states[-1] is None



"""Player abstractions for denon_rs232."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeAlias

from .const import (
    DigitalInputMode,
    InputSource,
    ModeSetting,
    RoomEQ,
    SurroundBack,
    TunerBand,
    TunerMode,
)
from .protocol import (
    channel_volume_to_param,
    parse_volume_param,
    volume_to_param,
)
from .state import MainZoneState, ZoneState

if TYPE_CHECKING:
    from .receiver import DenonReceiver


class _BasePlayer:
    """Shared stateful control surface for the main receiver and zones."""

    def __init__(
        self,
        receiver: DenonReceiver,
        state: ZoneState,
        *,
        power_command: str,
        power_standby_parameter: str,
        input_source_command: str,
        volume_command: str,
        half_db_volume_step: bool = True,
    ) -> None:
        self._receiver = receiver
        self._state = state
        self._power_command = power_command
        self._power_standby_parameter = power_standby_parameter
        self._input_source_command = input_source_command
        self._volume_command = volume_command
        self._half_db_volume_step = half_db_volume_step

    @property
    def power(self) -> bool | None:
        """Return the current power state for this player."""
        return self._state.power

    @property
    def input_source(self) -> InputSource | None:
        """Return the current input source for this player."""
        return self._state.input_source

    @property
    def volume(self) -> float | None:
        """Return the current volume for this player."""
        return self._state.volume

    @property
    def volume_min(self) -> float | None:
        """Return the shared minimum volume."""
        return self._receiver._state.main_zone.volume_min

    @property
    def volume_max(self) -> float | None:
        """Return the shared maximum volume."""
        return self._receiver._state.main_zone.volume_max

    async def power_on(self) -> None:
        """Turn this player on."""
        await self._receiver._send_command(self._power_command, "ON")

    async def power_standby(self) -> None:
        """Turn this player off/standby."""
        await self._receiver._send_command(
            self._power_command,
            self._power_standby_parameter,
        )

    async def select_input_source(self, source: InputSource) -> None:
        """Select an input source for this player."""
        await self._receiver._send_command(
            self._input_source_command,
            source.value,
        )

    async def volume_up(self) -> None:
        """Increase this player's volume."""
        await self._receiver._send_command(self._volume_command, "UP")

    async def volume_down(self) -> None:
        """Decrease this player's volume."""
        await self._receiver._send_command(self._volume_command, "DOWN")

    async def set_volume(self, db: float) -> None:
        """Set this player's volume in dB."""
        await self._receiver._send_command(
            self._volume_command,
            volume_to_param(db, self._half_db_volume_step),
        )

    async def query_power(self) -> bool:
        """Query the power state."""
        resp = await self._receiver._query(self._power_command)
        return resp == "ON"


class MainPlayer(_BasePlayer):
    """Stateful control surface for the receiver's main output."""

    _state: MainZoneState

    def __init__(self, receiver: DenonReceiver, state: MainZoneState) -> None:
        super().__init__(
            receiver,
            state,
            power_command="ZM",
            power_standby_parameter="OFF",
            input_source_command="SI",
            volume_command="MV",
            half_db_volume_step=True,
        )

    @property
    def mute(self) -> bool | None:
        """Return the current mute state."""
        return self._state.mute

    async def mute_on(self) -> None:
        """Mute the main player."""
        await self._receiver._send_command("MU", "ON")

    async def mute_off(self) -> None:
        """Unmute the main player."""
        await self._receiver._send_command("MU", "OFF")

    async def query_volume(self) -> float:
        """Query the current master volume."""
        resp = await self._receiver._query("MV")
        return parse_volume_param(resp)

    async def channel_volume_up(self, channel: str) -> None:
        """Adjust a channel volume up."""
        await self._receiver._send_command("CV", f"{channel} UP")

    async def channel_volume_down(self, channel: str) -> None:
        """Adjust a channel volume down."""
        await self._receiver._send_command("CV", f"{channel} DOWN")

    async def set_channel_volume(self, channel: str, db: float) -> None:
        """Set a channel volume in dB."""
        await self._receiver._send_command(
            "CV",
            f"{channel} {channel_volume_to_param(db)}",
        )

    async def query_mute(self) -> bool:
        """Query the current mute state."""
        resp = await self._receiver._query("MU")
        return resp == "ON"

    async def query_input_source(self) -> InputSource:
        """Query the active input source."""
        return InputSource(await self._receiver._query("SI"))

    async def set_surround_mode(self, mode: str) -> None:
        """Set surround mode."""
        await self._receiver._send_command("MS", mode)

    async def query_surround_mode(self) -> str:
        """Query the current surround mode."""
        return await self._receiver._query("MS")

    async def tone_defeat_on(self) -> None:
        """Enable tone defeat."""
        await self._receiver._send_command("PS", "TONE DEFEAT ON")

    async def tone_defeat_off(self) -> None:
        """Disable tone defeat."""
        await self._receiver._send_command("PS", "TONE DEFEAT OFF")

    async def set_surround_back(self, mode: SurroundBack) -> None:
        """Set surround back mode."""
        await self._receiver._send_command("PS", f"SB:{mode.value}")

    async def cinema_eq_on(self) -> None:
        """Enable cinema EQ."""
        await self._receiver._send_command("PS", "CINEMA EQ.ON")

    async def cinema_eq_off(self) -> None:
        """Disable cinema EQ."""
        await self._receiver._send_command("PS", "CINEMA EQ.OFF")

    async def set_mode_setting(self, mode: ModeSetting) -> None:
        """Set decoder mode."""
        await self._receiver._send_command("PS", f"MODE : {mode.value}")

    async def set_room_eq(self, mode: RoomEQ) -> None:
        """Set room EQ."""
        await self._receiver._send_command("PS", f"ROOM EQ:{mode.value}")

    async def set_digital_input(self, mode: DigitalInputMode) -> None:
        """Set digital input mode."""
        await self._receiver._send_command("SD", mode.value)

    async def query_digital_input(self) -> DigitalInputMode | None:
        """Query the current digital input mode."""
        param = await self._receiver._query("SD")
        if param == "NO":
            return None
        return DigitalInputMode(param)

    async def set_video_select(self, source: InputSource) -> None:
        """Set video select source."""
        await self._receiver._send_command("SV", source.value)

    async def cancel_video_select(self) -> None:
        """Cancel video select."""
        await self._receiver._send_command("SV", "SOURCE")

    async def query_video_select(self) -> InputSource | None:
        """Query the active video select source."""
        param = await self._receiver._query("SV")
        if param in ("SOURCE", "OFF"):
            return None
        return InputSource(param)

    async def set_rec_select(self, source: InputSource) -> None:
        """Set recording source."""
        await self._receiver._send_command("SR", source.value)

    async def cancel_rec_select(self) -> None:
        """Cancel recording source selection."""
        await self._receiver._send_command("SR", "SOURCE")

    async def query_rec_select(self) -> InputSource:
        """Query the current recording source."""
        return InputSource(await self._receiver._query("SR"))

    async def tuner_frequency_up(self) -> None:
        """Increment tuner frequency."""
        await self._receiver._send_command("TF", "UP")

    async def tuner_frequency_down(self) -> None:
        """Decrement tuner frequency."""
        await self._receiver._send_command("TF", "DOWN")

    async def set_tuner_frequency(self, freq: str) -> None:
        """Set tuner frequency directly."""
        await self._receiver._send_command("TF", freq)

    async def query_tuner_frequency(self) -> str:
        """Query the current tuner frequency."""
        return await self._receiver._query("TF")

    async def tuner_preset_up(self) -> None:
        """Increment tuner preset."""
        await self._receiver._send_command("TP", "UP")

    async def tuner_preset_down(self) -> None:
        """Decrement tuner preset."""
        await self._receiver._send_command("TP", "DOWN")

    async def set_tuner_preset(self, preset: str) -> None:
        """Set tuner preset directly."""
        await self._receiver._send_command("TP", preset)

    async def query_tuner_preset(self) -> str:
        """Query the current tuner preset."""
        return await self._receiver._query("TP")

    async def set_tuner_band(self, band: TunerBand) -> None:
        """Set tuner band."""
        await self._receiver._send_command("TM", band.value)

    async def set_tuner_mode(self, mode: TunerMode) -> None:
        """Set tuning mode."""
        await self._receiver._send_command("TM", mode.value)


class ZonePlayer(_BasePlayer):
    """Stateful control surface for a Denon zone."""

DenonPlayer: TypeAlias = MainPlayer | ZonePlayer



"""Known Denon receiver models and their capabilities."""

from __future__ import annotations

from dataclasses import dataclass

from .const import DigitalInputMode, InputSource


@dataclass(frozen=True)
class ReceiverModel:
    """Known capabilities of a Denon receiver model."""

    name: str
    input_sources: frozenset[InputSource]
    digital_inputs: frozenset[DigitalInputMode]
    surround_modes: tuple[str, ...]
    #: Zone 3 command prefix: "Z1" for legacy (AVR-3803/3805), "Z3" for modern.
    #: None means the model has no Zone 3.
    zone3_prefix: str | None = None
    #: Query prefixes to skip during connect() because the receiver does not
    #: answer them even though older models may.
    unsupported_startup_queries: frozenset[str] = frozenset()
    #: Whether Zone 2 or 3 supports 0.5 volume steps, as in "Z2805".
    #: False means the trailing "5" is stripped from the command: "Z280".
    zone_half_db_volume_step: bool = True


# -- Common source/digital sets used across multiple models --

_LEGACY_SOURCES = frozenset(
    {
        InputSource.PHONO,
        InputSource.CD,
        InputSource.TUNER,
        InputSource.DVD,
        InputSource.VDP,
        InputSource.TV,
        InputSource.DBS_SAT,
        InputSource.VCR_1,
        InputSource.VCR_2,
        InputSource.V_AUX,
        InputSource.CDR_TAPE1,
    }
)

_GEN1_DIGITAL = frozenset(
    {
        DigitalInputMode.AUTO,
        DigitalInputMode.PCM,
        DigitalInputMode.DTS,
        DigitalInputMode.ANALOG,
        DigitalInputMode.EXT_IN_1,
    }
)

_GEN1_DIGITAL_FULL = _GEN1_DIGITAL | frozenset(
    {
        DigitalInputMode.RF,
        DigitalInputMode.EXT_IN_2,
    }
)

_GEN2_DIGITAL = frozenset(
    {
        DigitalInputMode.AUTO,
        DigitalInputMode.HDMI,
        DigitalInputMode.DIGITAL,
        DigitalInputMode.ANALOG,
        DigitalInputMode.EXT_IN_1,
    }
)

_GEN3_DIGITAL = frozenset(
    {
        DigitalInputMode.AUTO,
        DigitalInputMode.HDMI,
        DigitalInputMode.DIGITAL,
        DigitalInputMode.ANALOG,
    }
)

# -- Common surround mode sets --

_COMMON_SURROUND = (
    "DIRECT",
    "PURE DIRECT",
    "STEREO",
    "MULTI CH IN",
    "MULTI CH DIRECT",
    "MULTI CH PURE D",
)

_DSP_SURROUND = (
    "WIDE SCREEN",
    "SUPER STADIUM",
    "ROCK ARENA",
    "JAZZ CLUB",
    "CLASSIC CONCERT",
    "MONO MOVIE",
    "MATRIX",
    "VIDEO GAME",
    "VIRTUAL",
)

_LEGACY_SURROUND = (
    *_COMMON_SURROUND,
    "DOLBY PRO LOGIC",
    "DOLBY PL2",
    "DOLBY PL2X",
    "DOLBY DIGITAL",
    "DOLBY D EX",
    "DTS NEO:6",
    "DTS SURROUND",
    "DTS ES DSCRT6.1",
    "DTS ES MTRX6.1",
    "5CH STEREO",
    "7CH STEREO",
    *_DSP_SURROUND,
)

_TRANSITION_SURROUND = (
    *_COMMON_SURROUND,
    "AUTO",
    "DOLBY PRO LOGIC",
    "DOLBY PL2",
    "DOLBY PL2X",
    "DOLBY DIGITAL",
    "DOLBY D EX",
    "DOLBY DIGITAL+",
    "DOLBY HD",
    "DTS NEO:6",
    "DTS SURROUND",
    "DTS ES DSCRT6.1",
    "DTS ES MTRX6.1",
    "DTS96/24",
    "MCH STEREO",
    "5CH STEREO",
    "7CH STEREO",
    *_DSP_SURROUND,
)

_MODERN_SURROUND = (
    *_COMMON_SURROUND,
    "AUTO",
    "DOLBY DIGITAL",
    "DOLBY DIGITAL+",
    "DOLBY HD",
    "DOLBY SURROUND",
    "DTS SURROUND",
    "DTS HD",
    "DTS HD MSTR",
    "MCH STEREO",
    *_DSP_SURROUND,
)


# -- Legacy era (~2003-2005) --

AVR_3803 = ReceiverModel(
    name="AVR-3803 / AVC-3570 / AVR-2803",
    input_sources=_LEGACY_SOURCES
    | frozenset(
        {
            InputSource.VCR_3,
            InputSource.MD_TAPE2,
        }
    ),
    digital_inputs=_GEN1_DIGITAL_FULL,
    surround_modes=_LEGACY_SURROUND,
    zone3_prefix="Z1",
)

AVR_3805 = ReceiverModel(
    name="AVR-3805 / AVC-3890",
    input_sources=_LEGACY_SOURCES,
    digital_inputs=_GEN1_DIGITAL,
    surround_modes=_LEGACY_SURROUND,
    zone3_prefix="Z1",
    zone_half_db_volume_step=False,
)

AVR_987 = ReceiverModel(
    name="AVR-987",
    input_sources=_LEGACY_SOURCES
    | frozenset(
        {
            InputSource.HDP,
            InputSource.DVR,
            InputSource.TV_CBL,
            InputSource.NET_USB,
        }
    ),
    digital_inputs=_GEN1_DIGITAL,
    surround_modes=_LEGACY_SURROUND,
    zone3_prefix="Z3",
)


# -- Transition era (~2006-2009) --

AVR_2308CI = ReceiverModel(
    name="AVR-2308CI / AVC-2308",
    input_sources=_LEGACY_SOURCES
    | frozenset(
        {
            InputSource.HDP,
            InputSource.DVR,
            InputSource.TV_CBL,
        }
    ),
    digital_inputs=_GEN1_DIGITAL,
    surround_modes=_TRANSITION_SURROUND,
)

AVR_2808CI = ReceiverModel(
    name="AVR-2808CI / AVC-2808 / AVR-988",
    input_sources=_LEGACY_SOURCES
    | frozenset(
        {
            InputSource.HDP,
            InputSource.DVR,
            InputSource.TV_CBL,
            InputSource.NET_USB,
        }
    ),
    digital_inputs=_GEN1_DIGITAL,
    surround_modes=_TRANSITION_SURROUND,
    zone3_prefix="Z3",
)

AVR_4308CI = ReceiverModel(
    name="AVR-4308CI",
    input_sources=_LEGACY_SOURCES
    | frozenset(
        {
            InputSource.HDP,
            InputSource.DVR,
            InputSource.TV_CBL,
            InputSource.NET_USB,
            InputSource.DOCK,
            InputSource.HDRADIO,
            InputSource.XM,
            InputSource.IPOD,
        }
    ),
    digital_inputs=_GEN1_DIGITAL,
    surround_modes=_TRANSITION_SURROUND,
    zone3_prefix="Z3",
)

AVR_3310CI = ReceiverModel(
    name="AVR-3310CI / AVR-990 / AVC-3310",
    input_sources=frozenset(
        {
            InputSource.PHONO,
            InputSource.CD,
            InputSource.TUNER,
            InputSource.DVD,
            InputSource.TV,
            InputSource.SAT_CBL,
            InputSource.DVR,
            InputSource.HDP,
            InputSource.V_AUX,
            InputSource.NET_USB,
            InputSource.DOCK,
            InputSource.HDRADIO,
            InputSource.IPOD,
        }
    ),
    digital_inputs=_GEN2_DIGITAL,
    surround_modes=_TRANSITION_SURROUND,
    zone3_prefix="Z3",
)


# -- Modern era (~2012+) --

AVR_X1000 = ReceiverModel(
    name="AVR-X1000 / AVR-E300",
    input_sources=frozenset(
        {
            InputSource.CD,
            InputSource.TUNER,
            InputSource.DVD,
            InputSource.BD,
            InputSource.TV,
            InputSource.SAT_CBL,
            InputSource.MPLAY,
            InputSource.GAME,
            InputSource.V_AUX,
            InputSource.AUX1,
            InputSource.NET,
            InputSource.USB_IPOD,
            InputSource.PANDORA,
            InputSource.SIRIUSXM,
            InputSource.SPOTIFY,
            InputSource.FLICKR,
            InputSource.IRADIO,
            InputSource.SERVER,
            InputSource.FAVORITES,
            InputSource.LASTFM,
        }
    ),
    digital_inputs=_GEN3_DIGITAL,
    surround_modes=_MODERN_SURROUND,
)

AVR_X4000 = ReceiverModel(
    name="AVR-X4000",
    input_sources=frozenset(
        {
            InputSource.PHONO,
            InputSource.CD,
            InputSource.TUNER,
            InputSource.DVD,
            InputSource.BD,
            InputSource.TV,
            InputSource.SAT_CBL,
            InputSource.MPLAY,
            InputSource.GAME,
            InputSource.V_AUX,
            InputSource.AUX1,
            InputSource.AUX2,
            InputSource.NET,
            InputSource.BT,
            InputSource.USB_IPOD,
            InputSource.PANDORA,
            InputSource.SIRIUSXM,
            InputSource.SPOTIFY,
            InputSource.FLICKR,
            InputSource.IRADIO,
            InputSource.SERVER,
            InputSource.FAVORITES,
            InputSource.LASTFM,
            InputSource.HDRADIO,
        }
    ),
    digital_inputs=_GEN3_DIGITAL,
    surround_modes=_MODERN_SURROUND,
    zone3_prefix="Z3",
)

AVR_X4200W = ReceiverModel(
    name="AVR-X4200W / AVR-X3200W / AVR-X2200W / AVR-X1200W",
    input_sources=frozenset(
        {
            InputSource.PHONO,
            InputSource.CD,
            InputSource.TUNER,
            InputSource.DVD,
            InputSource.BD,
            InputSource.TV,
            InputSource.SAT_CBL,
            InputSource.MPLAY,
            InputSource.GAME,
            InputSource.V_AUX,
            InputSource.AUX1,
            InputSource.AUX2,
            InputSource.NET,
            InputSource.BT,
            InputSource.USB_IPOD,
            InputSource.PANDORA,
            InputSource.SIRIUSXM,
            InputSource.SPOTIFY,
            InputSource.FLICKR,
            InputSource.IRADIO,
            InputSource.SERVER,
            InputSource.FAVORITES,
            InputSource.LASTFM,
            InputSource.HDRADIO,
        }
    ),
    digital_inputs=_GEN3_DIGITAL,
    surround_modes=(
        *_MODERN_SURROUND,
        "DOLBY ATMOS",
        "DTS:X",
        "DTS:X MSTR",
        "AURO3D",
        "AURO2DSURR",
    ),
    zone3_prefix="Z3",
)

AVR_X2700H = ReceiverModel(
    name="AVR-X2700H",
    input_sources=frozenset(
        {
            InputSource.PHONO,
            InputSource.CD,
            InputSource.TUNER,
            InputSource.DVD,
            InputSource.BD,
            InputSource.TV,
            InputSource.SAT_CBL,
            InputSource.MPLAY,
            InputSource.GAME,
            InputSource.AUX1,
            InputSource.AUX2,
            InputSource.NET,
            InputSource.BT,
            InputSource.USB_IPOD,
            InputSource.EIGHT_K,
        }
    ),
    digital_inputs=_GEN3_DIGITAL,
    surround_modes=(
        *_MODERN_SURROUND,
        "DOLBY ATMOS",
        "DTS:X",
        "DTS:X MSTR",
    ),
    unsupported_startup_queries=frozenset(
        {
            "SR",
            "TF",
            "TP",
        }
    ),
)

# -- Other (unknown model) --

OTHER = ReceiverModel(
    name="Other",
    input_sources=frozenset(InputSource),
    digital_inputs=frozenset(DigitalInputMode),
    surround_modes=(
        *_COMMON_SURROUND,
        "AUTO",
        "DOLBY PRO LOGIC",
        "DOLBY PL2",
        "DOLBY PL2X",
        "DOLBY DIGITAL",
        "DOLBY D EX",
        "DOLBY DIGITAL+",
        "DOLBY HD",
        "DOLBY SURROUND",
        "DOLBY ATMOS",
        "DTS NEO:6",
        "DTS SURROUND",
        "DTS ES DSCRT6.1",
        "DTS ES MTRX6.1",
        "DTS96/24",
        "DTS HD",
        "DTS HD MSTR",
        "DTS:X",
        "DTS:X MSTR",
        "MCH STEREO",
        "5CH STEREO",
        "7CH STEREO",
        *_DSP_SURROUND,
        "AURO3D",
        "AURO2DSURR",
    ),
)

#: All known receiver models, for iteration.
ALL_MODELS: tuple[ReceiverModel, ...] = (
    AVR_3803,
    AVR_3805,
    AVR_987,
    AVR_2308CI,
    AVR_2808CI,
    AVR_4308CI,
    AVR_3310CI,
    AVR_X1000,
    AVR_X4000,
    AVR_X4200W,
    AVR_X2700H,
)

#: Models keyed by identifier string, for lookup. Includes "other".
MODELS: dict[str, ReceiverModel] = {
    "avr_3803": AVR_3803,
    "avr_3805": AVR_3805,
    "avr_987": AVR_987,
    "avr_2308ci": AVR_2308CI,
    "avr_2808ci": AVR_2808CI,
    "avr_4308ci": AVR_4308CI,
    "avr_3310ci": AVR_3310CI,
    "avr_x1000": AVR_X1000,
    "avr_x4000": AVR_X4000,
    "avr_x4200w": AVR_X4200W,
    "avr_x2700h": AVR_X2700H,
    "other": OTHER,
}



"""Receiver implementation for denon_rs232."""

from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable
from typing import TYPE_CHECKING

import serialx

from .const import (
    BAUD_RATE,
    COMMAND_TIMEOUT,
    CR,
    DigitalInputMode,
    InputSource,
    ModeSetting,
    MULTI_RESPONSE_DELAY,
    POWER_ON_DELAY,
    PROBE_TIMEOUT,
    RoomEQ,
    SurroundBack,
    TunerBand,
    TunerMode,
    ZONE3_PREFIX,
    _MULTI_RESPONSE_PREFIXES,
    _SINGLE_RESPONSE_PREFIXES,
)
from .players import MainPlayer, ZonePlayer
from .protocol import (
    PendingQuery,
    _ZONE_VOL_RE,
    parse_channel_volume_param,
    parse_volume_param,
)
from .state import MainZoneState, ReceiverState, ZoneState

if TYPE_CHECKING:
    from .models import ReceiverModel

_LOGGER = logging.getLogger(__name__)


StateCallback = Callable[[ReceiverState | None], None]


class DenonReceiver:
    """Async controller for a Denon receiver over RS232."""

    def __init__(
        self,
        port: str,
        zone3_prefix: str = ZONE3_PREFIX,
        model: ReceiverModel | None = None,
        zone_half_db_volume_step: bool = True,
    ) -> None:
        self._port = port
        self._model = model
        self._zone3_prefix = (
            model.zone3_prefix or ZONE3_PREFIX if model else zone3_prefix
        )
        self._zone_half_db_volume_step = (
            model.zone_half_db_volume_step if model else zone_half_db_volume_step
        )
        self._reader: asyncio.StreamReader | None = None
        self._writer: serialx.SerialStreamWriter | None = None
        self._read_task: asyncio.Task | None = None
        self._power_on_task: asyncio.Task | None = None
        self._state = ReceiverState()
        self.main = MainPlayer(self, self._state.main_zone)
        self.zone_2 = ZonePlayer(
            self,
            self._state.zone_2,
            power_command="Z2",
            power_standby_parameter="OFF",
            input_source_command="Z2",
            volume_command="Z2",
            half_db_volume_step = self._zone_half_db_volume_step,
        )
        self.zone_3 = ZonePlayer(
            self,
            self._state.zone_3,
            power_command=self._zone3_prefix,
            power_standby_parameter="OFF",
            input_source_command=self._zone3_prefix,
            volume_command=self._zone3_prefix,
            half_db_volume_step = self._zone_half_db_volume_step,
        )
        self._subscribers: list[StateCallback] = []
        self._pending_queries: list[PendingQuery] = []
        self._write_lock = asyncio.Lock()
        self._connected = False
        self._batching = False
        self._batch_changed = False

    @property
    def model(self) -> ReceiverModel | None:
        """Return the receiver model, if set."""
        return self._model

    @property
    def state(self) -> ReceiverState:
        """Return a copy of the current state."""
        return self._state.copy()

    @property
    def connected(self) -> bool:
        return self._connected

    @property
    def power(self) -> bool | None:
        """Return the current receiver chassis power state."""
        return self._state.power

    def subscribe(self, callback: StateCallback) -> Callable[[], None]:
        """Subscribe to state changes. Returns an unsubscribe function."""
        self._subscribers.append(callback)
        return lambda: self._subscribers.remove(callback)

    async def connect(self) -> None:
        """Open the serial connection, verify, and query all state."""
        self._reader, self._writer = await serialx.open_serial_connection(
            self._port,
            baudrate=BAUD_RATE,
        )
        self._connected = True
        self._read_task = asyncio.create_task(self._read_loop())

        try:
            await self.query_power()
        except TimeoutError:
            await self.disconnect()
            raise ConnectionError(
                f"No response from receiver on {self._port}"
            ) from None

        _LOGGER.info("Connected to Denon receiver on %s", self._port)

    async def disconnect(self) -> None:
        """Close the serial connection."""
        await self._teardown()
        _LOGGER.info("Disconnected from Denon receiver")

    async def power_on(self) -> None:
        """Turn the receiver chassis on."""
        await self._send_command("PW", "ON")

    async def power_standby(self) -> None:
        """Put the receiver chassis in standby."""
        await self._send_command("PW", "STANDBY")

    async def query_power(self) -> bool:
        """Query the receiver chassis power state."""
        resp = await self._query("PW")
        if resp == "ON":
            return True
        if resp == "STANDBY":
            return False
        raise ValueError(f"Unknown power state: {resp}")

    async def query_state(self) -> None:
        """Query all initial state from the receiver.

        How much a receiver in standby answers differs per model, and every
        unanswered query costs a full COMMAND_TIMEOUT. While in standby the
        queries therefore stop at the first one that goes unanswered.

        Subscriber notifications are suppressed while the queries run and
        fired once at the end if any value changed.
        """
        unsupported_queries = (
            self._model.unsupported_startup_queries if self._model is not None else ()
        )
        self._cancel_power_on_query()
        self._batching = True
        self._batch_changed = False
        try:
            # The known power state can be stale, so refresh it before deciding
            # which queries the receiver is able to answer.
            try:
                await self._query("PW")
            except TimeoutError:
                pass
            standby = self._state.power is False

            for prefix in _SINGLE_RESPONSE_PREFIXES:
                if prefix == "PW" or prefix in unsupported_queries:
                    continue
                try:
                    await self._query(prefix)
                except TimeoutError:
                    if standby:
                        break

            # These are not waited on, so they stay worth sending in standby.
            for prefix in _MULTI_RESPONSE_PREFIXES:
                if prefix == "Z1":
                    if self._model is not None and self._model.zone3_prefix is None:
                        continue
                    prefix = self._zone3_prefix

                if prefix in unsupported_queries:
                    continue

                await self._send_command(prefix, "?")
                await asyncio.sleep(MULTI_RESPONSE_DELAY)
        finally:
            self._batching = False

        if self._batch_changed:
            self._notify_subscribers()

    async def probe_sources(
        self, timeout: float | None = None
    ) -> frozenset[InputSource]:
        """Probe which input sources the receiver supports."""
        if not self._connected:
            raise ConnectionError("Not connected")

        if timeout is None:
            timeout = PROBE_TIMEOUT

        original = self._state.main_zone.input_source
        available: set[InputSource] = set()

        if original is not None:
            available.add(original)

        for source in InputSource:
            if source == original:
                continue
            resp = await self._send_and_wait("SI", source.value, timeout=timeout)
            if resp == source.value:
                available.add(source)

        if original is not None:
            await self._send_and_wait("SI", original.value)

        return frozenset(available)

    async def _send_and_wait(
        self, prefix: str, param: str, timeout: float | None = None
    ) -> str | None:
        """Send a command and wait for a response with the given prefix."""
        if timeout is None:
            timeout = PROBE_TIMEOUT

        loop = asyncio.get_running_loop()
        future: asyncio.Future[str] = loop.create_future()
        pending = PendingQuery(prefix=prefix, future=future)
        self._pending_queries.append(pending)
        try:
            await self._send_command(prefix, param)
            return await asyncio.wait_for(future, timeout=timeout)
        except TimeoutError:
            return None
        finally:
            if pending in self._pending_queries:
                self._pending_queries.remove(pending)

    async def _send_command(self, command: str, parameter: str) -> None:
        """Send a command to the receiver."""
        assert self._writer is not None
        msg = f"{command}{parameter}\r".encode("ascii")
        _LOGGER.debug("Sending: %s", msg)
        try:
            async with self._write_lock:
                self._writer.write(msg)
                await self._writer.drain()
        except Exception:
            _LOGGER.exception("Error writing to serial port")
            await self._teardown()
            raise

    async def _query(self, command: str) -> str:
        """Send a query and wait for the response."""
        assert self._writer is not None
        loop = asyncio.get_running_loop()
        future: asyncio.Future[str] = loop.create_future()
        pending = PendingQuery(prefix=command, future=future)
        self._pending_queries.append(pending)

        try:
            msg = f"{command}?\r".encode("ascii")
            _LOGGER.debug("Querying: %s", msg)
            try:
                async with self._write_lock:
                    self._writer.write(msg)
                    await self._writer.drain()
            except Exception:
                _LOGGER.exception("Error writing to serial port")
                await self._teardown()
                raise
            return await asyncio.wait_for(future, timeout=COMMAND_TIMEOUT)
        finally:
            if pending in self._pending_queries:
                self._pending_queries.remove(pending)

    def _schedule_power_on_query(self) -> None:
        """Re-query state after the receiver leaves standby.

        Queries skipped while it was in standby are only answerable now, and
        the receiver needs a moment before it responds to them.
        """
        if self._power_on_task is not None and not self._power_on_task.done():
            return
        self._power_on_task = asyncio.create_task(self._delayed_query_state())

    def _cancel_power_on_query(self) -> None:
        """Drop a pending wake-up query.

        The wake-up query itself calls query_state(), which cancels here in
        turn, so the task must never cancel itself. The reference is kept
        until then so a teardown can stop a wake-up query that is running.
        """
        if self._power_on_task is None:
            return
        if self._power_on_task is not asyncio.current_task():
            self._power_on_task.cancel()
        self._power_on_task = None

    async def _delayed_query_state(self) -> None:
        """Wait for the receiver to wake up, then query its state."""
        await asyncio.sleep(POWER_ON_DELAY)
        try:
            await self.query_state()
        except Exception:
            _LOGGER.exception("Error querying state after power on")

    async def _teardown(self) -> None:
        """Tear down the connection after an error."""
        if not self._connected:
            return
        self._connected = False

        current = asyncio.current_task()

        self._cancel_power_on_query()

        if self._read_task is not None and self._read_task is not current:
            self._read_task.cancel()
            try:
                await self._read_task
            except asyncio.CancelledError:
                pass
        self._read_task = None

        if self._writer is not None:
            self._writer.close()
            await self._writer.wait_closed()
            self._writer = None
            self._reader = None

        self._notify_subscribers()

    async def _read_loop(self) -> None:
        """Continuously read and process messages from the receiver."""
        assert self._reader is not None
        buf = b""

        while self._connected:
            try:
                data = await self._reader.read(256)
            except Exception:
                if not self._connected:
                    return
                _LOGGER.exception("Error reading from serial port")
                await self._teardown()
                return

            if not data:
                _LOGGER.warning("Serial connection closed")
                await self._teardown()
                return

            buf += data

            while CR in buf:
                line, buf = buf.split(CR, 1)
                if not line:
                    continue
                message = line.decode("ascii", errors="replace").strip()
                if message:
                    self._process_message(message)

    @staticmethod
    def _set_attr_value(target: object, attr: str, new_value: object) -> bool:
        """Set an attribute only when the value changed."""
        if getattr(target, attr) == new_value:
            return False
        setattr(target, attr, new_value)
        return True

    def _set_state_value(self, attr: str, new_value: object) -> bool:
        """Set a MainZoneState attribute only when the value changed."""
        return self._set_attr_value(self._state.main_zone, attr, new_value)

    def _process_message(self, message: str) -> None:
        """Parse and process a message from the receiver."""
        _LOGGER.debug("Received: %s", message)

        if len(message) < 2:
            return

        prefix = message[:2]
        param = message[2:]

        if prefix == "MV" and param.startswith("MAX"):
            prefix = "MVMAX"
            param = param.removeprefix("MAX").strip()
        elif prefix == "MV" and param.startswith("MIN"):
            prefix = "MVMIN"
            param = param.removeprefix("MIN").strip()

        changed = False

        if prefix == "PW":
            if param == "ON":
                was_standby = self._state.power is False
                changed = self._set_attr_value(self._state, "power", True)
                if was_standby and not self._batching:
                    self._schedule_power_on_query()
            elif param == "STANDBY":
                changed = self._set_attr_value(self._state, "power", False)
            else:
                _LOGGER.warning("Unknown power state: %s", param)

        elif prefix == "ZM":
            changed = self._set_state_value("power", param == "ON")

        elif prefix == "MV":
            try:
                changed = self._set_state_value("volume", parse_volume_param(param))
            except (ValueError, IndexError):
                _LOGGER.warning("Could not parse volume: %s", param)

        elif prefix == "MVMAX":
            try:
                changed = self._set_state_value("volume_max", parse_volume_param(param))
            except (ValueError, IndexError):
                _LOGGER.warning("Could not parse max volume: %s", param)

        elif prefix == "MVMIN":
            try:
                changed = self._set_state_value("volume_min", parse_volume_param(param))
            except (ValueError, IndexError):
                _LOGGER.warning("Could not parse min volume: %s", param)

        elif prefix == "MU":
            changed = self._set_state_value("mute", param == "ON")

        elif prefix == "SI":
            try:
                changed = self._set_state_value("input_source", InputSource(param))
            except ValueError:
                _LOGGER.warning("Unknown input source: %s", param)

        elif prefix == "MS":
            changed = self._set_state_value("surround_mode", param)

        elif prefix == "CV":
            channel, sep, val = param.partition(" ")
            if sep and val not in ("UP", "DOWN"):
                try:
                    new_value = parse_channel_volume_param(val)
                    if self._state.main_zone.channel_volumes.get(channel) != new_value:
                        self._state.main_zone.channel_volumes[channel] = new_value
                        changed = True
                except (ValueError, IndexError):
                    _LOGGER.warning("Could not parse channel volume: %s", param)

        elif prefix == "PS":
            changed = self._process_ps_param(param)

        elif prefix == "SD":
            if param == "NO":
                changed = self._set_state_value("digital_input", None)
            else:
                try:
                    changed = self._set_state_value(
                        "digital_input",
                        DigitalInputMode(param),
                    )
                except ValueError:
                    _LOGGER.warning("Unknown digital input mode: %s", param)

        elif prefix == "SV":
            if param in ("SOURCE", "OFF"):
                changed = self._set_state_value("video_select", None)
            else:
                try:
                    changed = self._set_state_value("video_select", InputSource(param))
                except ValueError:
                    _LOGGER.warning("Unknown video source: %s", param)

        elif prefix == "SR":
            if param == "SOURCE":
                changed = self._set_state_value("rec_select", None)
            else:
                try:
                    changed = self._set_state_value("rec_select", InputSource(param))
                except ValueError:
                    _LOGGER.warning("Unknown rec source: %s", param)

        elif prefix == "TF":
            if param not in ("UP", "DOWN"):
                changed = self._set_state_value("tuner_frequency", param)

        elif prefix == "TP":
            if param not in ("UP", "DOWN"):
                changed = self._set_state_value("tuner_preset", param)

        elif prefix == "TM":
            try:
                changed = self._set_state_value("tuner_band", TunerBand(param))
            except ValueError:
                try:
                    changed = self._set_state_value("tuner_mode", TunerMode(param))
                except ValueError:
                    _LOGGER.warning("Unknown tuner setting: %s", param)

        elif prefix == "Z2":
            changed = self._process_zone_param(self._state.zone_2, param)

        elif prefix == "Z1":
            changed = self._process_zone_param(self._state.zone_3, param)

        elif prefix == "Z3":
            changed = self._process_zone_param(self._state.zone_3, param)

        for pending in list(self._pending_queries):
            if pending.prefix == prefix and not pending.future.done():
                pending.future.set_result(param)

        if changed:
            if self._batching:
                self._batch_changed = True
            else:
                self._notify_subscribers()

    def _process_ps_param(self, param: str) -> bool:
        """Process a PS (parameter setting) parameter."""
        if param == "TONE DEFEAT ON":
            return self._set_state_value("tone_defeat", True)
        if param == "TONE DEFEAT OFF":
            return self._set_state_value("tone_defeat", False)
        if param.startswith("SB:"):
            try:
                return self._set_state_value("surround_back", SurroundBack(param[3:]))
            except ValueError:
                _LOGGER.warning("Unknown surround back mode: %s", param)
                return False
        if param == "CINEMA EQ.ON":
            return self._set_state_value("cinema_eq", True)
        if param == "CINEMA EQ.OFF":
            return self._set_state_value("cinema_eq", False)
        if param.startswith("MODE : "):
            try:
                return self._set_state_value("mode_setting", ModeSetting(param[7:]))
            except ValueError:
                _LOGGER.warning("Unknown mode setting: %s", param)
                return False
        if param.startswith("ROOM EQ:"):
            try:
                return self._set_state_value("room_eq", RoomEQ(param[8:]))
            except ValueError:
                _LOGGER.warning("Unknown room EQ mode: %s", param)
                return False

        _LOGGER.debug("Unknown PS parameter: %s", param)
        return False

    def _process_zone_param(self, zone: ZoneState, param: str) -> bool:
        """Process a Z2/Z1/Z3 parameter."""
        if param == "ON":
            return self._set_attr_value(zone, "power", True)
        if param == "OFF":
            return self._set_attr_value(zone, "power", False)
        if param in ("UP", "DOWN"):
            return False
        if _ZONE_VOL_RE.match(param):
            try:
                return self._set_attr_value(zone, "volume", parse_volume_param(param))
            except (ValueError, IndexError):
                return False
        if param.startswith("SLP"):
            return False
        if param == "SOURCE":
            return self._set_attr_value(zone, "input_source", None)
        try:
            return self._set_attr_value(zone, "input_source", InputSource(param))
        except ValueError:
            _LOGGER.warning("Unknown zone source: %s", param)
            return False

    def _notify_subscribers(self) -> None:
        """Notify all subscribers of a state change or disconnect."""
        state = self._state.copy() if self._connected else None
        for callback in self._subscribers:
            try:
                callback(state)
            except Exception:
                _LOGGER.exception("Error in state change callback %s", callback)



"""Constants and enums shared across the denon_rs232 package."""

from enum import Enum

BAUD_RATE = 9600
COMMAND_TIMEOUT = 2.0  # seconds to wait for a response
MULTI_RESPONSE_DELAY = 0.3  # seconds to wait for multi-response query results
PROBE_TIMEOUT = 0.8  # seconds to wait for each probe attempt
POWER_ON_DELAY = 3.0  # seconds a receiver needs after PW ON before it answers
CR = b"\r"

# Volume range constants (dB)
MIN_VOLUME_DB = -80.0
MAX_VOLUME_DB = 18.0
VOLUME_DB_RANGE = MAX_VOLUME_DB - MIN_VOLUME_DB  # 98.0

# Prefixes that return a single response to "?", safe for _query().
_SINGLE_RESPONSE_PREFIXES = (
    "PW",
    "ZM",
    "MV",
    "MU",
    "SI",
    "MS",
    "SD",
    "SV",
    "SR",
    "TF",
    "TP",
)

# Prefixes that return multiple responses to "?", state populated via _process_message.
_MULTI_RESPONSE_PREFIXES = ("CV", "PS", "TM", "Z2", "Z1")

# Zone 3 prefix: legacy models (AVR-3803/3805) use "Z1", modern models use "Z3".
ZONE3_PREFIX = "Z3"


class InputSource(Enum):
    """Input sources available on the Denon receiver."""

    # Legacy sources (~2003-2007)
    PHONO = "PHONO"
    CD = "CD"
    TUNER = "TUNER"
    DVD = "DVD"
    VDP = "VDP"
    TV = "TV"
    DBS_SAT = "DBS/SAT"
    VCR_1 = "VCR-1"
    VCR_2 = "VCR-2"
    VCR_3 = "VCR-3"
    V_AUX = "V.AUX"
    CDR_TAPE1 = "CDR/TAPE1"
    MD_TAPE2 = "MD/TAPE2"

    # Transition era (~2006-2009)
    HDP = "HDP"
    DVR = "DVR"
    TV_CBL = "TV/CBL"
    SAT = "SAT"
    NET_USB = "NET/USB"
    DOCK = "DOCK"
    IPOD = "IPOD"

    # Modern era (~2012-2016)
    BD = "BD"
    SAT_CBL = "SAT/CBL"
    MPLAY = "MPLAY"
    GAME = "GAME"
    AUX1 = "AUX1"
    AUX2 = "AUX2"
    NET = "NET"
    BT = "BT"
    USB_IPOD = "USB/IPOD"
    EIGHT_K = "8K"

    # Streaming / online services
    PANDORA = "PANDORA"
    SIRIUSXM = "SIRIUSXM"
    SPOTIFY = "SPOTIFY"
    FLICKR = "FLICKR"
    IRADIO = "IRADIO"
    SERVER = "SERVER"
    FAVORITES = "FAVORITES"
    LASTFM = "LASTFM"

    # Radio services (region-specific)
    XM = "XM"
    SIRIUS = "SIRIUS"
    HDRADIO = "HDRADIO"
    DAB = "DAB"


class DigitalInputMode(Enum):
    """Digital input modes."""

    AUTO = "AUTO"
    # Gen 1 (legacy ~2003-2007)
    PCM = "PCM"
    DTS = "DTS"
    RF = "RF"
    ANALOG = "ANALOG"
    EXT_IN_1 = "EXT.IN-1"
    EXT_IN_2 = "EXT.IN-2"
    # Gen 2+ (~2009+)
    HDMI = "HDMI"
    DIGITAL = "DIGITAL"


class SurroundBack(Enum):
    """Surround back speaker modes."""

    MTRX_ON = "MTRX ON"
    NON_MTRX = "NON MTRX"
    PL2X_CINEMA = "PL2X CINEMA"
    PL2X_MUSIC = "PL2X MUSIC"
    OFF = "OFF"


class ModeSetting(Enum):
    """Decoder mode settings (for PL2/PL2x/NEO:6)."""

    MUSIC = "MUSIC"
    CINEMA = "CINEMA"
    GAME = "GAME"
    PRO_LOGIC = "PRO LOGIC"


class RoomEQ(Enum):
    """Room EQ modes."""

    NORMAL = "NORMAL"
    FRONT = "FRONT"
    FLAT = "FLAT"
    MANUAL = "MANUAL"
    OFF = "OFF"


class TunerBand(Enum):
    """Tuner bands."""

    AM = "AM"
    FM = "FM"


class TunerMode(Enum):
    """Tuning modes."""

    AUTO = "AUTO"
    MANUAL = "MANUAL"



"""Protocol helpers for denon_rs232."""

from __future__ import annotations

import asyncio
import re
from dataclasses import dataclass

_ZONE_VOL_RE = re.compile(r"^\d{2,3}$")


def parse_volume_param(param: str) -> float:
    """Convert a master volume parameter string to a dB value."""
    if param == "99":
        return -80.0

    if len(param) == 3:
        whole = int(param[:2])
        return (whole - 80) + 0.5
    return int(param) - 80


def volume_to_param(db: float, half_step: bool = True) -> str:
    """Convert a dB value to the master volume protocol parameter."""
    if db <= -80:
        return "99"

    raw = db + 80
    whole = int(raw)
    if not half_step:
        return f"{whole:02d}"
    if raw - whole >= 0.5:
        return f"{whole:02d}5"
    return f"{whole:02d}"


def parse_channel_volume_param(param: str) -> float:
    """Convert a channel volume parameter string to a dB value."""
    if len(param) == 3:
        whole = int(param[:2])
        return (whole - 50) + 0.5
    return int(param) - 50


def channel_volume_to_param(db: float) -> str:
    """Convert a dB value to the channel volume protocol parameter."""
    raw = db + 50
    whole = int(raw)
    if raw - whole >= 0.5:
        return f"{whole:02d}5"
    return f"{whole:02d}"


@dataclass
class PendingQuery:
    """A pending query waiting for a response."""

    prefix: str
    future: asyncio.Future[str]



"""Runtime state dataclasses for denon_rs232."""

from __future__ import annotations

from dataclasses import dataclass, field, replace

from .const import (
    DigitalInputMode,
    InputSource,
    ModeSetting,
    RoomEQ,
    SurroundBack,
    TunerBand,
    TunerMode,
)


@dataclass
class ZoneState:
    """State for Zone 2 or Zone 3."""

    power: bool | None = None
    input_source: InputSource | None = None
    volume: float | None = None

    def copy(self) -> ZoneState:
        return replace(self)


@dataclass
class MainZoneState(ZoneState):
    """State for the main listening zone.

    Extends ZoneState (power, input_source, volume) with main-zone-only
    fields such as mute, surround mode, channel volumes, and tuner state.
    """

    mute: bool | None = None
    volume_max: float | None = None
    volume_min: float | None = None
    #: Surround mode name. Kept as str because the receiver returns many
    #: combined mode names (e.g. "DOLBY D+PL2X C", "M CH IN+PL2X M").
    surround_mode: str | None = None

    # Channel volumes. Event-only. Keyed by channel: FL, FR, C, SW, SL, SR, SBL, SBR, SB.
    channel_volumes: dict[str, float] = field(default_factory=dict)

    # Parameter settings. Event-only. Populated from PS? responses and PS events.
    tone_defeat: bool | None = None
    surround_back: SurroundBack | None = None
    cinema_eq: bool | None = None
    mode_setting: ModeSetting | None = None
    #: Event-only. The receiver does not include room_eq in PS? responses.
    room_eq: RoomEQ | None = None

    # Digital / video / rec (queryable)
    digital_input: DigitalInputMode | None = None
    video_select: InputSource | None = None
    rec_select: InputSource | None = None

    # Tuner (frequency and preset are queryable; band and mode are event-only)
    tuner_frequency: str | None = None
    tuner_preset: str | None = None
    #: Event-only. Updated from TM events.
    tuner_band: TunerBand | None = None
    #: Event-only. Updated from TM events. AUTO or MANUAL.
    tuner_mode: TunerMode | None = None

    def copy(self) -> MainZoneState:
        return replace(self, channel_volumes=dict(self.channel_volumes))


@dataclass
class ReceiverState:
    """Current state of the Denon receiver."""

    power: bool | None = None
    main_zone: MainZoneState = field(default_factory=MainZoneState)
    zone_2: ZoneState = field(default_factory=ZoneState)
    zone_3: ZoneState = field(default_factory=ZoneState)

    def copy(self) -> ReceiverState:
        return replace(
            self,
            main_zone=self.main_zone.copy(),
            zone_2=replace(self.zone_2),
            zone_3=replace(self.zone_3),
        )



"""Tests for denon_rs232 receiver model definitions."""

from denon_rs232 import DigitalInputMode, InputSource
from denon_rs232.models import (
    ALL_MODELS,
    AVR_2308CI,
    AVR_2808CI,
    AVR_3310CI,
    AVR_3803,
    AVR_3805,
    AVR_4308CI,
    AVR_987,
    AVR_X1000,
    AVR_X2700H,
    AVR_X4000,
    AVR_X4200W,
)


# -- AVR-3805 (legacy) --


def test_avr_3805_excludes_invalid_sources():
    """AVR-3805 should not include VCR-3 or MD/TAPE2."""
    assert InputSource.VCR_3 not in AVR_3805.input_sources
    assert InputSource.MD_TAPE2 not in AVR_3805.input_sources


def test_avr_3805_includes_valid_sources():
    assert InputSource.CD in AVR_3805.input_sources
    assert InputSource.DVD in AVR_3805.input_sources
    assert InputSource.DBS_SAT in AVR_3805.input_sources


def test_avr_3805_excludes_invalid_digital():
    """AVR-3805 should not include RF or EXT.IN-2."""
    assert DigitalInputMode.RF not in AVR_3805.digital_inputs
    assert DigitalInputMode.EXT_IN_2 not in AVR_3805.digital_inputs


def test_avr_3805_includes_valid_digital():
    assert DigitalInputMode.AUTO in AVR_3805.digital_inputs
    assert DigitalInputMode.EXT_IN_1 in AVR_3805.digital_inputs


def test_avr_3805_zone3_prefix():
    assert AVR_3805.zone3_prefix == "Z1"


def test_avr_3805_does_not_support_half_db_volume_steps():
    assert AVR_3805.zone_half_db_volume_step == False


# -- AVR-3803 (legacy with full digital set) --


def test_avr_3803_includes_vcr3_and_md():
    """AVR-3803 has VCR-3 and MD/TAPE2."""
    assert InputSource.VCR_3 in AVR_3803.input_sources
    assert InputSource.MD_TAPE2 in AVR_3803.input_sources


def test_avr_3803_includes_rf_and_ext2():
    """AVR-3803 has Gen1 full digital set including RF and EXT.IN-2."""
    assert DigitalInputMode.RF in AVR_3803.digital_inputs
    assert DigitalInputMode.EXT_IN_2 in AVR_3803.digital_inputs


def test_avr_3803_zone3_prefix():
    assert AVR_3803.zone3_prefix == "Z1"


# -- AVR-2308CI (transition, no zone 3) --


def test_avr_2308ci_no_zone3():
    assert AVR_2308CI.zone3_prefix is None


def test_avr_2308ci_has_hdp():
    assert InputSource.HDP in AVR_2308CI.input_sources


def test_avr_2308ci_no_modern_sources():
    assert InputSource.BD not in AVR_2308CI.input_sources
    assert InputSource.MPLAY not in AVR_2308CI.input_sources


# -- AVR-3310CI (transition to modern digital) --


def test_avr_3310ci_has_hdmi_digital():
    """AVR-3310CI uses Gen2 digital inputs with HDMI and DIGITAL."""
    assert DigitalInputMode.HDMI in AVR_3310CI.digital_inputs
    assert DigitalInputMode.DIGITAL in AVR_3310CI.digital_inputs


def test_avr_3310ci_no_legacy_digital():
    """AVR-3310CI should not have PCM/DTS/RF individual digital modes."""
    assert DigitalInputMode.PCM not in AVR_3310CI.digital_inputs
    assert DigitalInputMode.RF not in AVR_3310CI.digital_inputs


def test_avr_3310ci_zone3():
    assert AVR_3310CI.zone3_prefix == "Z3"


# -- AVR-X1000 (modern, no zone 3) --


def test_avr_x1000_modern_sources():
    assert InputSource.BD in AVR_X1000.input_sources
    assert InputSource.MPLAY in AVR_X1000.input_sources
    assert InputSource.GAME in AVR_X1000.input_sources
    assert InputSource.SAT_CBL in AVR_X1000.input_sources


def test_avr_x1000_streaming():
    assert InputSource.SPOTIFY in AVR_X1000.input_sources
    assert InputSource.PANDORA in AVR_X1000.input_sources


def test_avr_x1000_no_zone3():
    assert AVR_X1000.zone3_prefix is None


def test_avr_x1000_no_legacy_sources():
    assert InputSource.VDP not in AVR_X1000.input_sources
    assert InputSource.DBS_SAT not in AVR_X1000.input_sources
    assert InputSource.VCR_1 not in AVR_X1000.input_sources


# -- AVR-X4000 (modern with zone 3 and BT) --


def test_avr_x4000_has_bt():
    assert InputSource.BT in AVR_X4000.input_sources


def test_avr_x4000_zone3():
    assert AVR_X4000.zone3_prefix == "Z3"


def test_avr_x4000_gen3_digital():
    assert DigitalInputMode.AUTO in AVR_X4000.digital_inputs
    assert DigitalInputMode.HDMI in AVR_X4000.digital_inputs
    assert DigitalInputMode.DIGITAL in AVR_X4000.digital_inputs
    assert DigitalInputMode.ANALOG in AVR_X4000.digital_inputs
    assert DigitalInputMode.EXT_IN_1 not in AVR_X4000.digital_inputs


# -- AVR-X4200W (newest) --


def test_avr_x4200w_has_all_modern():
    assert InputSource.BT in AVR_X4200W.input_sources
    assert InputSource.HDRADIO in AVR_X4200W.input_sources
    assert InputSource.SPOTIFY in AVR_X4200W.input_sources


def test_avr_x4200w_zone3():
    assert AVR_X4200W.zone3_prefix == "Z3"


# -- AVR-X2700H (modern with 8K source) --


def test_avr_x2700h_has_8k_and_modern_core_sources():
    assert InputSource.EIGHT_K in AVR_X2700H.input_sources
    assert InputSource.BD in AVR_X2700H.input_sources
    assert InputSource.MPLAY in AVR_X2700H.input_sources
    assert InputSource.NET in AVR_X2700H.input_sources


def test_avr_x2700h_no_zone3():
    assert AVR_X2700H.zone3_prefix is None


def test_avr_x2700h_skips_known_unsupported_startup_queries():
    assert AVR_X2700H.unsupported_startup_queries == {"SR", "TF", "TP"}


# -- General model checks --


def test_all_models_tuple():
    """ALL_MODELS should contain all defined models."""
    assert len(ALL_MODELS) == 11
    assert AVR_3805 in ALL_MODELS
    assert AVR_X4200W in ALL_MODELS
    assert AVR_X2700H in ALL_MODELS


def test_all_models_have_auto_digital():
    """Every model should support AUTO digital input mode."""
    for model in ALL_MODELS:
        assert DigitalInputMode.AUTO in model.digital_inputs, model.name


def test_all_models_have_cd():
    """Every model should support CD input."""
    for model in ALL_MODELS:
        assert (
            InputSource.CD in model.digital_inputs
            or InputSource.CD in model.input_sources
        ), model.name
